import os
import collections
import time
import json
import grpc
import logging
import base64
from . import generic_client_interceptor

from . import user_pb2 as user

from .user_pb2_grpc import UserServiceStub
from .org_pb2_grpc import OrgServiceStub
from .project_pb2_grpc import ProjectServiceStub
from .dataset_pb2_grpc import DatasetServiceStub
from .file_pb2_grpc import FileServiceStub
from .app_pb2_grpc import AppServiceStub
from .feature_set_pb2_grpc import FeatureSetServiceStub
from .model_pb2_grpc import ModelServiceStub
from .quota_manager_pb2_grpc import QuotaManagerServiceStub
from .template_pb2_grpc import TemplateServiceStub
from .pipeline_pb2_grpc import PipelineServiceStub
from .recipe_pb2_grpc import RecipeServiceStub
from .config_pb2_grpc import ConfigServiceStub
from .data_source_pb2_grpc import DataSourceServiceStub
from .preprocessing_pipeline_pb2_grpc import PreprocessingPipelineServiceStub
from .artifact_pb2_grpc import ArtifactServiceStub

class _ClientCallDetails(
        collections.namedtuple(
            '_ClientCallDetails',
            ('method', 'timeout', 'metadata', 'credentials')),
        grpc.ClientCallDetails):
    pass

class Client():

    def __init__(self,
                 target: str = None,
                 secure: bool = None,
                 custom_token: str = None,
                 email: str = None,
                 password: str = None,
                 config_file_path: str = None):
        if not target:
            target = os.getenv('AIA_ENGINE_API_TARGET', '')
        if secure is None:
            secure = os.getenv('AIA_ENGINE_API_SECURE', 'true') == 'true'
        if not custom_token:
            custom_token = os.getenv('AIA_ENGINE_API_CUSTOM_TOKEN', '')
        if not email:
            email = os.getenv('AIA_ENGINE_API_EMAIL', '')
        if not password:
            password = os.getenv('AIA_ENGINE_API_PASSWORD', '')
        if not config_file_path:
            config_file_path = os.environ.get('AIA_ENGINE_CONFIG_FILE', 'aiaengine_config.json')

        #Find from env before from config file
        if target != '' and (custom_token != '' or (email != '' and password != '')):
            provider = 'token' if custom_token != '' else 'password'
            data = {'token': custom_token} if custom_token != '' else {'email': email, 'password': password}
            self.config = {
                'target': target,
                'secure': secure,
                'auth': {
                    'provider': provider,
                    'data': data
                }
            }
        else:
            logging.debug(f"authenticate using AIA_ENGINE_CONFIG_FILE: {config_file_path}")
            with open(config_file_path, 'r') as f:
                self.config = json.load(f)
        target = self.config['target']
        logging.debug(f"connecting to {target}")
        secure = self.config['secure']

        max_msg_length = 10 * 1024 * 1024
        options = [('grpc.max_message_length', max_msg_length),
                   ('grpc.max_send_message_length', max_msg_length),
                   ('grpc.max_receive_message_length', max_msg_length)]

        if secure:
            credentials = grpc.ssl_channel_credentials()
            channel = grpc.secure_channel(target, credentials=credentials, options=options)
        else:
            channel = grpc.insecure_channel(target, options=options)
        #user service doesn't require refresh_token
        self._no_interceptor_user_svc = UserServiceStub(channel)

        authentication_interceptor = generic_client_interceptor.create(self._intercept_call)
        self.channel = grpc.intercept_channel(channel, authentication_interceptor)

        self.token = ''
        self.refresh_token = None
        self.token_expired_at = 0

        self._user_service = UserServiceStub(self.channel)
        self._org_service = OrgServiceStub(self.channel)
        self._project_service = ProjectServiceStub(self.channel)
        self._dataset_service = DatasetServiceStub(self.channel)
        self._file_service = FileServiceStub(self.channel)
        self._app_service = AppServiceStub(self.channel)
        self._featureset_service = FeatureSetServiceStub(self.channel)
        self._model_service = ModelServiceStub(self.channel)
        self._quota_manager_sevice = QuotaManagerServiceStub(self.channel)
        self._template_service = TemplateServiceStub(self.channel)
        self._pipeline_service = PipelineServiceStub(self.channel)
        self._recipe_service = RecipeServiceStub(self.channel)
        self._config_service = ConfigServiceStub(self.channel)
        self._datasource_service = DataSourceServiceStub(self.channel)
        self._preprocessing_pipeline_service = PreprocessingPipelineServiceStub(self.channel)
        self._artifact_service = ArtifactServiceStub(self.channel)

        self._authenticate()


    def _authenticate(self):
        authentication_config = self.config['auth']
        authentication_provider = authentication_config['provider']
        if authentication_provider == 'password' or authentication_provider == 'email_password':
            email = authentication_config['data']['email']
            password = authentication_config['data']['password']
            response_data = self._no_interceptor_user_svc.SignInWithEmail(user.SignInWithEmailRequest(email=email, password=password))
            self.token = response_data.token
            self.refresh_token = response_data.refresh_token
            self.token_expired_at = time.time() + int(response_data.expires_in)
        elif authentication_provider == 'token':
            token = authentication_config['data']['token']
            response_data = self._no_interceptor_user_svc.SignInWithCustomToken(user.SignInWithCustomTokenRequest(token=token))
            self.token = response_data.token
            self.refresh_token = response_data.refresh_token
            self.token_expired_at = time.time() + int(response_data.expires_in)

        if self.token is not None:
            self.user_id = json.loads(base64.b64decode(self.token.split('.')[1] + '==').decode('ascii'))['sub']

    def _refresh_token_if_needed(self):
        if self.token_expired_at > time.time():
            return
        logging.debug('refreshing access token')
        response_data = self._no_interceptor_user_svc.RefreshToken(user.RefreshTokenRequest(refresh_token=self.refresh_token))
        self.token = response_data.token
        self.refresh_token = response_data.refresh_token
        self.token_expired_at = time.time() + int(response_data.expires_in)

    @property
    def users(self):
        return self._user_service

    @property
    def orgs(self):
        return self._org_service

    @property
    def projects(self):
        return self._project_service

    @property
    def datasets(self):
        return self._dataset_service

    @property
    def files(self):
        return self._file_service

    @property
    def apps(self):
        return self._app_service

    @property
    def featuresets(self):
        return self._featureset_service

    @property
    def models(self):
        return self._model_service

    @property
    def quota_manager(self):
        return self._quota_manager_sevice

    @property
    def templates(self):
        return self._template_service

    @property
    def pipelines(self):
        return self._pipeline_service

    @property
    def recipes(self):
        return self._recipe_service
    
    @property
    def configs(self):
        return self._config_service

    @property
    def datasources(self):
        return self._datasource_service

    @property
    def preprocessing_pipelines(self):
        return self._preprocessing_pipeline_service

    @property
    def artifacts(self):
        return self._artifact_service

    def close(self):
        self.channel.close()

    def _intercept_call(self, client_call_details, request_iterator, request_streaming,
                        response_streaming):
        self._refresh_token_if_needed()
        metadata = []
        if client_call_details.metadata is not None:
            metadata = list(client_call_details.metadata)
        metadata.append((
            'authorization',
            'Bearer ' + self.token,
        ))
        client_call_details = _ClientCallDetails(
            client_call_details.method, client_call_details.timeout, metadata,
            client_call_details.credentials)
        return client_call_details, request_iterator, None
