import json
from dataclasses import dataclass, field
from enum import Enum
from typing import List, Set, Dict, Union
from dataclasses_json import dataclass_json, config

import grpc
from google.protobuf import json_format, struct_pb2
from aiaengine.api import LATEST_VERSION, model as api_model
from aiaengine.api import featureset as api_feature_set
from aiaengine.api import app as api_app
from aiaengine.dataset import Dataset
from aiaengine.common import RECOMMENDED_FEATURE_SET, DEFAULT_TIMEOUT, Resource
from aiaengine.core.model import HyperparameterTuningMethod
from aiaengine.feature_set import FeatureSet
from aiaengine.model import Model
from aiaengine.artifact import Artifact, ArtifactType
from aiaengine.core import (
    MLProblemType,
    ClassificationSubType,
    TrainingConfig
)
from aiaengine.api import artifact as api_artifact

class TimeUnit(str, Enum):
    DAY = 'day'
    WEEK = 'week'
    MONTH = 'month'

@dataclass_json
@dataclass
class Duration:
    unit: TimeUnit
    value: int

class FeatureType(str, Enum):
    TEXT = 'Text'
    NUMERIC = 'Numeric'
    CATEGORICAL = 'Categorical'

@dataclass_json
@dataclass(frozen=True)
class AutoModelSelection:
    top_n: int = 1

@dataclass_json
@dataclass(frozen=True)
class ManualModelSelection:
    template_id: str
    name: str
    training_config: TrainingConfig = None

@dataclass_json
@dataclass(frozen=True)
class AutoFeatureSelection:
    threshold: float = 0.5

@dataclass_json
@dataclass(frozen=True)
class ManualFeatureSelection:
    name: str
    selected_features: Set[str]
    description: str = ""
    preferred_feature_types: Dict[str, FeatureType] = None

class TrainTestSplitMethod(str, Enum):
    RANDOM = "random"
    TIME_BASED = "time_based"

@dataclass_json
@dataclass(frozen=True)
class TrainTestSplitConfig:
    method: TrainTestSplitMethod = TrainTestSplitMethod.RANDOM
    train_percentage: float = 0.8
    test_portion_time_window: Duration = None
    timestamp_column: str = None

def model_selection_encoder(value):
    if value is None:
        return None

    if isinstance(value, AutoModelSelection):
        return {
            'auto': json.loads(value.to_json())
        }
    else:
        return {
            'manual': [json.loads(item.to_json()) for item in value]
        }

def feature_selection_encoder(value):
    if value is None:
        return None

    if isinstance(value, AutoFeatureSelection):
        return {
            'auto': json.loads(value.to_json())
        }
    elif isinstance(value, ManualFeatureSelection):
        return {
            'manual': json.loads(value.to_json())
        }

@dataclass_json
@dataclass(frozen=True)
class SupervisedLearningConfig:
    target_column: str
    """The models will be trained to predict a value of the target column"""
    train_test_split: TrainTestSplitConfig = TrainTestSplitConfig()
    """Configure the train/test split"""

    feature_selection: Union[AutoFeatureSelection, ManualFeatureSelection] = field(
        metadata=config(
            encoder=feature_selection_encoder,
        ),
        default=AutoFeatureSelection()
    )
    """Configure the features used to train the models"""

    model_selection: Union[AutoModelSelection, List[ManualModelSelection]] = field(
        metadata=config(
            encoder=model_selection_encoder,
        ),
        default=AutoModelSelection()
    )
    """Configure which models to train"""

@dataclass_json
@dataclass(frozen=True)
class ClassificationConfig(SupervisedLearningConfig):
    """Configuration for Classification application
    """

    sub_type: ClassificationSubType = ClassificationSubType.MULTI_CLASS
    """Sub-type of the classification problem: Binary or Multi-class classification
    """

    positive_class_label: str = None
    """If sub_type is Binary, this must be the value of positive class
    """

    negative_class_label: str = None
    """If sub_type is Binary, this must be the value of negative class 
    """

@dataclass_json
@dataclass(frozen=True)
class RegressionConfig(SupervisedLearningConfig):
    """Configuration for Regression application
    """

    pass

@dataclass_json
@dataclass(frozen=True)
class ClusteringConfig:
    """Configuration for Clustering application

        Example:

            ```Python
            # create a clustering application
            app = project.create_app(
                name='Insurance Customer Segmentation - Test 8',
                dataset_id=dataset.id,
                problem_type=MLProblemType.CLUSTERING,
                problem_type_config=ClusteringConfig(
                    feature_set={
                        'selected_features': [
                            "State",
                            "Response",
                            "Coverage",
                            "Education",
                            "EmploymentStatus",
                            "Gender",
                            "Income",
                            "Location Code",
                            "Marital Status",
                            "Monthly Premium Auto",
                            "Months Since Last Claim",
                            "Months Since Policy Inception",
                            "Number of Open Complaints",
                            "Number of Policies",
                            "Policy Type",
                            "Policy",
                            "Renew Offer Type",
                            "Sales Channel",
                            "Total Claim Amount",
                            "Vehicle Class",
                            "Vehicle Size",
                        ]
                    },
                    initial_models=[
                        {
                            'name': 'GMM',
                            'template_id': Clusterers.GMM,
                            'training_config':{
                                'hyperparameter_tuning_method': HyperparameterTuningMethod.GRID_SEARCH,
                                'hyperparameter_tuning_config': None,
                                'hyperparameters': {
                                    'k': {
                                        'max_num_clusters': 5,
                                        'min_num_clusters': 2
                                    }
                                },
                            }
                            
                        },
                        {
                            'name': 'K-Means',
                            'template_id': Clusterers.KMeans,
                            'training_config': {
                                'hyperparameter_tuning_method': None,
                                'hyperparameter_tuning_config': None,
                                'hyperparameters': {},
                            }
                        }
                    ]
                ),
                training_data_proportion=1.0
            )
            ```
    """

    feature_set: dict = None
    """Selected features
    """
    initial_models: list = None
    """List of initial models will be trained when the app is created
    """

@dataclass_json
@dataclass(frozen=True)
class AnomalyDetectionConfig:
    """Configuration for Anomaly Detection application
    """

    pass

@dataclass_json
@dataclass(frozen=True)
class TimeSeriesForecastingConfig:
    """Configuration for Time-series Forecasting application
    """

    time_index_column: str
    target_columns: Set[str]


class App(Resource):
    """The object model represents an Application in the Engine
    """

    @property
    def name(self) -> str:
        """Name of the app

        Returns:
            str: Name of the app
        """
        return self._data.name

    @property
    def description(self) -> str:
        """Description of the app

        Returns:
            str: Description of the app
        """
        return self._data.description

    def update(self, name: str, description: str = None):
        """Update information of the app

        Args:
            name (str): updated project name
            description (str, optional): updated project description. Defaults to None.
        """
        self.client.apps.UpdateApp(
            api_app.UpdateAppRequest(
                id=self.id,
                name=name,
                description=description
            )
        )

    def delete(self):
        """Delete the app.

            Warning:
                All related data (models, feature sets, deployments) of the app will be deleted.
        """
        self.client.apps.DeleteApp(
            api_app.DeleteAppRequest(
                id=self.id
            )
        )

    def create_feature_set(
        self,
        name: str,
        feature_names: Set[str],
        description: str = None,
        timeout: int = DEFAULT_TIMEOUT,
    ) -> FeatureSet:
        """Create a new feature set

        Args:
            name (str): name of the feature set. Name must be unique in the Application-scope.
            feature_names (Set[str]): Name of columns used as features
            description (str, optional): description of the feature set. Defaults to None.
            timeout (int, optional): waiting time in seconds. Defaults to DEFAULT_TIMEOUT.

        Raises:
            Exception: _description_

        Returns:
            FeatureSet: the newly created `FeatureSet` resource
        """
        try:
            feature_set_data = self.client.featuresets.CreateFeatureSet(
                api_feature_set.CreateFeatureSetRequest(
                    name=name,
                    description=description,
                    app_id=self.id,
                    selected_features=feature_names,
                )
            )
        except grpc.RpcError as e:
            raise Exception(e.details())
        feature_set = FeatureSet(
            id=feature_set_data.id, data=feature_set_data, client=self.client
        )

        # wait_for_processing(self.client, "featureset", response.id, timeout)
        feature_set.wait_for_processed(timeout=timeout)

        return feature_set

    def list_feature_sets(self) -> List[FeatureSet]:
        """Get all available feature sets

        Returns:
            List of FeatureSet resources
        """
        response = self.client.featuresets.ListFeatureSets(
            api_feature_set.ListFeatureSetsRequest(app_id=self.id)
        )
        return [
            FeatureSet(
                id=feature_set_data.id, data=feature_set_data, client=self.client
            )
            for feature_set_data in response.feature_sets
        ]

    def get_feature_set(
        self, id: str = None, timeout: int = DEFAULT_TIMEOUT,
    ) -> FeatureSet:
        """Get feature set by ID

        Args:
            id (str, optional): ID of the feature set. Defaults to None.
            timeout (int, optional): waiting for the feature set processed time in seconds. Defaults to DEFAULT_TIMEOUT.

        Returns:
            FeatureSet: _description_
        """
        feature_set_data = self.client.featuresets.GetFeatureSet(
            api_feature_set.GetFeatureSetRequest(id=id)
        )
        feature_set = FeatureSet(
            id=feature_set_data.id, data=feature_set_data, client=self.client,
        )
        feature_set.wait_for_processed(timeout=timeout)

    def get_recommended_feature_set(self, timeout: int = DEFAULT_TIMEOUT) -> FeatureSet:
        """Get the recommended feature set of the application

        Args:
            timeout (int, optional): time (in seconds) to wait for feature set processed. Defaults to DEFAULT_TIMEOUT.

        Raises:
            RuntimeError: raised if the recommended feature set is not available at the time

        Returns:
            return Recommended feature set of the application
        """
        feature_sets = list(
            filter(lambda f: f.name == RECOMMENDED_FEATURE_SET, self.list_feature_sets())
        )
        if len(feature_sets) == 0:
            raise RuntimeError("Recommended feature set is not available yet")
        recommended_feature_set = feature_sets[0]
        recommended_feature_set.wait_for_processed(timeout=timeout)
        return recommended_feature_set

    def create_model(
        self,
        name: str,
        template_id: str,
        feature_set_id: str,
        dataset_version: int = LATEST_VERSION,
        hyperparameters: dict = None,
        hyperparameter_tuning_method: HyperparameterTuningMethod = None,
        hyperparameter_tuning_config: dict = None,
        continuous_learning: bool = False,
        pipeline_configuration_id: str = None,
        wait_until_trained: bool = True,
        timeout: int = DEFAULT_TIMEOUT * 2,
    ) -> Model:
        """Train a new model using a specific `dataset_version` of the dataset

        Args:
            name (str): descriptive name of the model
            template_id (str): the ID of the pre-defined model template
            feature_set_id (str): ID of the feature set used to train the model
            dataset_version (int, optional): _description_. Defaults to `LATEST_VERSION`.
            continuous_learning (bool, optional): _description_. Defaults to False.
            pipeline_configuration_id (str, optional): _description_. Defaults to None.
            wait_until_trained (bool, optional): _description_. Defaults to True.
            timeout (int, optional): _description_. Defaults to DEFAULT_TIMEOUT*2.

        Returns:
            Model: the newly created model
        """
        struct_hyperparameters = json_format.ParseDict(hyperparameters, struct_pb2.Struct()) if hyperparameters else None
        struct_hyperparameter_tuning_config = json_format.ParseDict(hyperparameter_tuning_config, struct_pb2.Struct()) if hyperparameter_tuning_config else None
        api_training_config = api_model.TrainingConfig(
            hyperparameters=struct_hyperparameters,
            hyperparameter_tuning_method=hyperparameter_tuning_method,
            hyperparameter_tuning_config=struct_hyperparameter_tuning_config
        ) if struct_hyperparameters or struct_hyperparameter_tuning_config or hyperparameter_tuning_method else None

        model_data = self.client.models.CreateModel(
            api_model.CreateModelRequest(
                app_id=self.id,
                feature_set_id=feature_set_id,
                dataset_version=dataset_version,
                name=name,
                template_id=template_id,
                training_config=api_training_config,
                continuous_learning=continuous_learning,
                pipeline_configuration_id=pipeline_configuration_id,
            )
        )

        model = Model(id=model_data.id, data=model_data, client=self.client)

        if wait_until_trained:
            # wait_for_processing(self.client, "model", model_id, timeout)
            model.wait_for_trained(dataset_version, timeout)

        return model

    def list_models(self) -> List[Model]:
        """Get all available models

        Returns:
            List[Model]: list of Model resources
        """
        response = self.client.models.ListModels(
            api_model.ListModelsRequest(app_id=self.id)
        )
        return [
            Model(id=model_data.id, data=model_data, client=self.client)
            for model_data in response.models
        ]

    def list_artifacts(self, type: str, labels: Dict[str, str] = None) -> List[Artifact]:
        """List all artifacts belong to this application which match the type and the labels

        Args:
            type (str): type of the artifact
            labels (Dict[str, str]): labels of the artifacts. Defaults to None.

        Returns:
            List of artifacts
        """
        labels = {} if labels is None else labels
        labels['app'] = self.id
        labels['project'] = self._data.project.id
        labels['org'] = self._data.org.id
        response = self.client.artifacts.ListArtifacts(
            api_artifact.ListArtifactsRequest(
                type=type,
                labels=labels
            )
        )

        return [
            Artifact(id=artifact.id, data=artifact, client=self.client)
            for artifact in response.artifacts
        ]

    def _get_data(self) -> any:
        return self.client.apps.GetApp(api_app.GetAppRequest(id=self.id))

@dataclass
class ClusterInsight:
    name: str
    data: dict

@dataclass
class ClusteringResult:
    name: str
    output_dataset: Dataset
    insights: List[ClusterInsight]

def _get_clustering_result_from_model(app: App, model: Model) -> ClusteringResult:
    artifacts = app.list_artifacts(
        type=ArtifactType.ClusteringResult,
        labels={
            'model': model.id
        }
    )
    if len(artifacts) == 0:
        return None

    output_dataset = Dataset(id=artifacts[0].data['output_dataset'])
    insight_artifacts = app.list_artifacts(
        type=ArtifactType.ClusterInsights,
        labels={
            'model': model.id
        },
    )
    insights = [
        ClusterInsight(name=artifact.metadata.name, data=artifact.data)
        for artifact in insight_artifacts
    ]
    return ClusteringResult(
        name=model.name,
        output_dataset=output_dataset,
        insights=insights
    )

def get_clustering_result(app: App, name: str) -> ClusteringResult:
    """Get a clustering result of the app by name

    Args:
        app (App): The clustering app
        name (str): Name of the result

    Raises:
        Exception: Result not found

    Returns:
        ClusteringResult: the clustering result
    """
    models = app.list_models()
    model = None
    for m in models:
        if m.name == name:
            model = m
            break
    if model is None:
        raise Exception(f"Result with name {name} not found in the app.")

    return _get_clustering_result_from_model(app, model)

def get_all_clustering_results(app: App) -> List[ClusteringResult]:
    """Get all clustering results of the app

    Args:
        app (App): The clustering app

    Returns:
        List[ClusteringResult]: Clustering results
    """
    models = app.list_models()
    # wait for all results available
    [model.wait_for_trained() for model in models]

    results = []
    for model in models:
        result = _get_clustering_result_from_model(app, model)
        if result is not None:
            results.append(result)

    return results

@dataclass
class ClusteringApp:
    app: App

    def get_result(self, name: str) -> ClusteringResult:
        return get_clustering_result(app=self.app, name=name)

    def get_all_results(self) -> List[ClusteringResult]:
        return get_all_clustering_results(app=self.app)