from typing import List, Union, Dict
from enum import Enum
from dataclasses import dataclass, field
from dataclasses_json import dataclass_json
from aiaengine.core.dataset import Dataset
class MLProblemType(str, Enum):
    CLASSIFICATION = 'classification'
    REGRESSION = 'regression'
    CLUSTERING = 'clustering'
    ANOMALY_DETECTION = 'anomaly-detection'
    TIME_SERIES_FORECASTING = 'forecasting'

class ClassificationSubType(str, Enum):
    BINARY = 'binary'
    MULTI_CLASS = 'multi-class'

@dataclass(frozen=True)
class BaseProblemTypeConfig:
    template_config: dict = None

@dataclass(frozen=True)
class SupervisedLearningConfig(BaseProblemTypeConfig):
    target_column: str = None
    """The models will be trained to predict a value of the target column"""
    model_selection: dict = None
    feature_selection: dict = None
    train_test_split: dict = None
    train_data_proportion: float = 0.8
    """The propotion of input dataset used for training
    the remaining part will be used as a test set to evaluate the models
    """

@dataclass(frozen=True)
class ClassificationConfig(SupervisedLearningConfig):
    sub_type: ClassificationSubType = ClassificationSubType.MULTI_CLASS
    positive_class_label: str = None
    negative_class_label: str = None

    @staticmethod
    def from_dict(dict: dict):
        return ClassificationConfig(
            template_config=dict.get('template_config', None),
            target_column=dict.get('target_column', None),
            train_data_proportion=dict.get('train_data_proportion', 0.8),
            sub_type=dict.get('sub_type', ClassificationSubType.MULTI_CLASS),
            positive_class_label=dict.get('positive_class_label', None),
            negative_class_label=dict.get('negative_class_label', None),
            model_selection=dict.get('model_selection', None),
            feature_selection=dict.get('feature_selection', None),
            train_test_split=dict.get('train_test_split', None)
        )

@dataclass(frozen=True)
class RegressionConfig(SupervisedLearningConfig):
    @staticmethod
    def from_dict(dict: dict):
        return RegressionConfig(
            template_config=dict.get('template_config', None),
            target_column=dict.get('target_column', None),
            train_data_proportion=dict.get('train_data_proportion', 0.8),
            model_selection=dict.get('model_selection', None),
            feature_selection=dict.get('feature_selection', None),
            train_test_split=dict.get('train_test_split', None)
        )

@dataclass(frozen=True)
class ClusteringConfig(BaseProblemTypeConfig):
    feature_set: dict = None
    initial_models: list = None

    @staticmethod
    def from_dict(dict: dict):
        return ClusteringConfig(
            template_config=dict.get('template_config', None),
            feature_set=dict.get('feature_set', None),
            initial_models=dict.get('initial_models', None),
        )

@dataclass(frozen=True)
class AnomalyDetectionConfig(BaseProblemTypeConfig):
    @staticmethod
    def from_dict(dict: dict):
        return AnomalyDetectionConfig(
            template_config=dict.get('template_config', None),
        )

@dataclass(frozen=True)
class TimeSeriesForecastingConfig(BaseProblemTypeConfig):
    time_index_column: str = None

    @staticmethod
    def from_dict(dict: dict):
        return TimeSeriesForecastingConfig(
            template_config=dict.get('template_config', None),
            time_index_column=dict.get('time_index_column', None),
        )

@dataclass(frozen=True)
class App:
    name: str
    problem_type: MLProblemType
    target_columns: List[str] # Deprecated: use problem_type_config.target_column
    problem_type_config: Union[ClassificationConfig, 
                               RegressionConfig,
                               ClusteringConfig,
                               AnomalyDetectionConfig,
                               TimeSeriesForecastingConfig] = None
    time_index_column: str = None # Deprecated: use problem_type_config.time_index_column
    description: str = None
    dataset: Dataset = None
    train_data: Dataset = None
    test_data: Dataset = None
    id: str = None
    project_id: str = None
    dataset_version: int = -1
    org_id: str = None

    annotations: Dict[str, str] = field(default_factory=dict)
    status: str = ""