from dataclasses import dataclass
import json
from typing import IO, List, Union
from aiaengine.io_utils import load_file_to_json

DEFAULT_SAMPLE_SIZE = 5000  # default sample size

@dataclass
class Column:
    """Column information
    """
    name: str
    data_type: str

    def to_dict(self) -> dict:
        return {
            'name': self.name,
            'data_type': self.data_type
        }

@dataclass(frozen=True)
class Schema:
    """Schema information of the dataset
    """
    columns: List[Column]

    def clone(self):
        return Schema(columns=self.columns)

    def to_dict(self):
        return {'columns': [column.to_dict() for column in self.columns]}

    @staticmethod
    def from_file(file: Union[str,IO]):
        if isinstance(file, str):
            return Schema.from_dict(load_file_to_json(file))
        else:
            return Schema.from_dict(json.load(file))

    @staticmethod
    def from_dict(data: dict):
        return Schema(columns=[Column(name=c['name'], data_type=c['data_type']) for c in data['columns']])

class Dataset:
    """An object model represents a Dataset in the platform
    """
    def __init__(self, schema: Schema, data_urls: List[str], id: str=None, version: int=None,
                metadata: dict=None,
                first_n_rows_sample_dataset=None,
                random_n_rows_sample_dataset=None):
        """
        Args:
            schema (Schema | dict | file-like): [description]
            data_urls (list of urls): [description]
            id: indentity of dataset in the platform if saved
            version: dataset version
            metadata: analysis metadata
            first_n_rows_sample_dataset: a Dataset of first N rows
            random_n_rows_sample_dataset: a Dataset of random sample data
        """
        self.id = id
        self.version = version
        self.schema = schema
        self.data_urls = data_urls
        self.metadata = metadata
        # this 2 Datasets for sample data
        self._first_n_rows_sample_dataset = first_n_rows_sample_dataset
        self._random_n_rows_sample_dataset = random_n_rows_sample_dataset

    @property
    def first_n_rows_sample(self):
        """ First N row sample """
        return self._first_n_rows_sample_dataset

    @property
    def random_n_rows_sample(self):
        """ Random sample """
        return self._random_n_rows_sample_dataset
