from typing import List, Set
from enum import Enum
from dataclasses import dataclass, replace
from aiaengine.core.feature_set import FeatureSet
from aiaengine.core.dataset import Dataset
from aiaengine.util import download_model
import tempfile

MODEL_FORMAT_SCIKIT_LEARN = 'scikit-learn'
MODEL_FORMAT_XGBOOST = 'XGBoost'
MODEL_FORMAT_LIGHTGBM = 'LightGBM'
MODEL_FORMAT_ONNX = 'ONNX'
MODEL_FORMAT_CUML = 'cuML'
MODEL_FORMAT_SPARKML = 'SparkML'
class HyperparameterTuningMethod(str, Enum):
    NONE = None
    GRID_SEARCH = 'GridSearch'
    RANDOM_SEARCH = 'RandomSearch'
    BAYESIAN_SEARCH = 'BayesianSearch'

@dataclass(frozen=True)
class Model:
    """Represent the model asset in the platform

    Args:
        name (str): name of the model
        template_id (str): The model template from which this model was trained
        feature_set (FeatureSet): The feature set used to train the model
        id (str, optional): Unique identity of the model in the platform. Defaults to None.
    """
    name: str
    template_id: str
    feature_set: FeatureSet
    id: str = None
    dataset_version: int = -1
    continuous_learning: bool = False

@dataclass
class TrainingConfig:
    """Training configuration"""
    hyperparameter_tuning_method: HyperparameterTuningMethod
    hyperparameter_tuning_config: dict
    hyperparameters: dict
    """
        hyperparameter_tuning_method (HyperparameterTuningMethod): Grid, Random, or Bayesian?
        hyperparameter_tuning_config (dict): Optimization configuration of the tuning method
        hyperparameters (dict): Initial value of algorithm's hyperparameters or the grid
    """

@dataclass(frozen=True)
class TrainedModel:
    format: str
    url: str
    hyperparameters: dict = None
    model: any = None
    output_columns: Set[str] = None

    def download(self):
        """Download model to local environment and return new instance
        Returns:
            [TrainedModel]: Local instance
        """
        model_path = tempfile.mkdtemp()
        download_model(self.url, model_path)
        return replace(self, url=model_path)

@dataclass(frozen=True)
class Training:
    """Configuration of a specific model training

    Args:
        config (TrainingConfig): training configuration
        id (str, optional): [description]. Defaults to None.
    """
    model: Model
    config: TrainingConfig
    id: str = None
    trained_model: TrainedModel = None
    dataset_version: int = -1

@dataclass(frozen=True)
class TrainingResult:
    """Contains the training results (best parameters...), statistics
    """
    training: Training
    trained_model: TrainedModel
    training_stats: dict

@dataclass
class EvaluationConfig:
    """The configuration of an evaluation
    """
    grouping_column: str = None

@dataclass(frozen=True)
class Evaluation:
    training: Training
    config: EvaluationConfig
    test_dataset: Dataset # Evaluating dataset
    evaluation_stats: dict = None
    id: str = None

@dataclass(frozen=True)
class EvaluationResult:
    """Result of the evaluation
    """
    evaluation: Evaluation
    summary: dict
    details: dict = None

@dataclass(frozen=True)
class PredictionConfig:
    include_variables = False

@dataclass(frozen=True)
class Prediction:
    training: Training
    config: PredictionConfig
    input_dataset: Dataset = None
    id: str = None

@dataclass(frozen=True)
class PredictionResult:
    prediction: Prediction
    output_dataset: Dataset = None
