from dataclasses import dataclass
from enum import Enum
import os
from typing import List, Tuple
from xmlrpc.client import boolean
import requests
import logging
import re
from google.protobuf import json_format
from aiaengine.api import (
    LATEST_VERSION,
    data_source,
    file,
    pipeline,
    Client,
)

from aiaengine.common import DEFAULT_TIMEOUT, STATUS_FAILED, STATUS_SUCCESS
from aiaengine.common import Resource
from aiaengine.waiter import wait_until, wait_pipeline_run


class FileType(str, Enum):
    """Supported file types
    """
    CSV = "csv"
    """CSV file"""
    Parquet = "parquet"
    """Parquet file"""
    JSONLine = "json"
    """JSON lines file"""
    Excel = "excel"
    """Excel file"""


class DataType(Enum):
    # this class define the data_type name
    Numeric = "Numeric"
    Boolean = "Boolean"
    DateTime = "DateTime"
    Text = "Text"


class FileSettings:
    def asdict(self):
        return {}


class CSVSettings(FileSettings):
    """Settings for CSV files
    """
    delimiter = ","
    escape = '"'
    quote = '"'
    header = True
    date_format = "yyyy-MM-dd HH:mm:ss"
    ignore_leading_whitespace = True
    ignore_trailing_whitespace = True
    multiline = True

    def asdict(self):
        return {
            "delimiter": self.delimiter,
            "escape": self.escape,
            "quote": self.quote,
            "header": self.header,
            "dateFormat": self.date_format,
            "ignoreLeadingWhiteSpace": self.ignore_leading_whitespace,
            "ignoreTrailingWhiteSpace": self.ignore_trailing_whitespace,
            "multiLine": self.multiline,
        }


@dataclass
class ExcelSettings(FileSettings):
    """Settings for Excel files
    """
    data_range: str
    header: bool = True

    def asdict(self):
        return {
            "header": self.header,
            "cellsRange": self.data_range
        }

@dataclass
class Column:
    """Column definition
    """
    name: str
    type: DataType
    renamed: str = None

    def rename(self, renamed):
        self.renamed = renamed
        return self


class SchemaBuilder:
    name_field = "name"
    data_type_field = "data_type"
    alias_field = "alias"

    def __init__(self, columns: List[Column]):
        self._columns = []
        self._used_columns_name = []
        if columns is not None:
            for col in columns:
                self.add_column(
                    col.name, col.type, col.renamed,
                )

    def add_column(self, name, data_type: DataType, renamed=None):
        new_col = {self.name_field: name, self.data_type_field: data_type.value}
        self._columns.append(new_col)
        if isinstance(renamed, str):
            renamed = renamed.strip()
        else:
            renamed = ""
        if len(renamed) > 0:
            if name != renamed:
                new_col[self.alias_field] = renamed
            self._used_columns_name.append(renamed)
        # else:
        #     renamed = self._rename_according_convention(name)
        #     if name != renamed:
        #         logging.debug(f"automatically rename column: {name} => {renamed}")
        #         new_col[self.alias_field] = renamed
        #     self._used_columns_name.append(renamed)
        return self

    def update_column(self, name, data_type: DataType = None, renamed=None):
        for i, d in enumerate(self._columns):
            if d[self.name_field] == name:
                if data_type is not None:
                    d[self.data_type_field] = data_type.value
                if renamed is not None:
                    d[self.alias_field] = renamed
                    self._used_columns_name[i] = renamed
        return self

    def _rename_according_convention(self, original_column):
        #   AI & A Engine will automatically populate the New Column Name, by making necessary changes to original column name that does not meet the requirements:
        # 	Must be unique (names are case insensitive): Two columns named "col_1" and "Col_1" -> The latter duplicated column will have an underscore added as a suffix "Col_1_".
        # 	Must start with a letter: To add the prefix "Col_" for column names that do not start with a letter.
        # 	Must not be an empty string: Stick to current implementation. "Unnamed_XX". XX represents the column number.
        # 	Can only contain letters, numbers, and underscores:  All special characters not supported will be suppressed to an underscore "_".
        # 	Have to ensure that column names do not collide AFTER suppressing them according to the rules above: Two columns named '10$000' and '10,000', after the substitution we will end up with 10_000 twice. So we have to add another underscore to the second one, 10_000 and 10_000_.
        rename_column = re.sub(r"[^A-Za-z0-9_]+", "_", original_column)
        # column start with number, replaced by prefix Col_
        if rename_column[0] >= "0" and rename_column[0] <= "9":
            rename_column = "Col_" + rename_column
        elif rename_column[0] == "_":
            rename_column = "Col" + rename_column

        # ensure no two columns have same name
        while rename_column in self._used_columns_name:
            rename_column = rename_column + "_"
        return rename_column

    def build(self):
        return {"columns": self._columns}


class SettingsBuilder:
    def __init__(self, settings=None):
        self._settings = {}
        if settings is not None:
            self._file_type = settings["fileType"]
            self._settings = settings["settings"]

    def set_file_type(self, file_type):
        self._file_type = file_type
        return self

    def update_settings(self, key, value):
        # settings: settings of the file {"comment": "", "dateFormat": "yyyy-MM-dd HH:mm:ss", "delimiter": ",", "encoding": "UTF-8", "escape": "\"", "header": true,
        #              "ignoreLeadingWhiteSpace": true, "ignoreTrailingWhiteSpace": true, "multiLine": true, "quote": "\""}

        self._settings[key] = value

    def build(self):
        return {"fileType": self._file_type, "settings": self._settings}


class DataSource(Resource):
    # Underlying API DataSource object
    data: data_source.DataSource
    file_settings: dict
    schema: dict

    def _create_connection_info(self, project_id: str) -> Tuple[str, dict]:
        pass

    def create(self, project_id: str, name: str, timeout: int = DEFAULT_TIMEOUT):
        source_type, connection_info = self._create_connection_info(project_id)

        create_data_source_response = self.client.datasources.CreateDataSource(
            json_format.ParseDict(
                {
                    "name": name,
                    "project_id": project_id,
                    "source_type": source_type,
                    "connection_info": connection_info,
                },
                data_source.CreateDataSourceRequest(),
            )
        )

        data_source_id = create_data_source_response.data_source.id
        job_id = create_data_source_response.job_id
        # if upload file from local, and it's not Excel file, we don't run pipeline
        if job_id:
            wait_pipeline_run(self.client, job_id, timeout)

        def data_source_processed():
            self.data = self.client.datasources.GetDataSource(
                data_source.GetDataSourceRequest(id=data_source_id)
            )
            self.id = self.data.id
            if self.data.status == "data_ready":
                return True
            if self.data.status == STATUS_FAILED:
                raise RuntimeError(f"Failed to create DataSource(id={self.data.id})")

        wait_until(data_source_processed, timeout=timeout)

        # update file settings and schema
        update_file_settings = {
            "id": self.data.id,
            "shared_file_settings": self.file_settings,
        }
        self.client.datasources.UpdateFileSettings(
            json_format.ParseDict(
                update_file_settings, data_source.UpdateFileSettingsRequest()
            )
        )

        update_data_item_schema = {
            "id": self.data.id,
            "shared_schema": self.schema,
        }
        self.client.datasources.UpdateDataItemSchema(
            json_format.ParseDict(
                update_data_item_schema, data_source.UpdateDataItemSchemaRequest()
            )
        )

        return self.data

    def update(self, timeout: int = DEFAULT_TIMEOUT):
        logging.debug(f"Getting DataSource(id={self.id})")
        current_ds = self.client.datasources.GetDataSource(
            data_source.GetDataSourceRequest(id=self.id)
        )

        source_type, connection_info = self._create_connection_info(
            current_ds.project.id
        )
        if source_type != current_ds.source_type:
            raise RuntimeError(
                f"Source type mismatched(original = {current_ds.source_type}, updated source type = {source_type}"
            )

        # logging.debug(f"Getting DataSource(id={self.id})")
        # current_ds = self.client.datasources.GetDataSource(data_source.GetDataSourceRequest(id=self.id))
        # current_urls = json_format.MessageToDict(current_ds.connection_info["urls"])
        # urls = list(set(current_urls + urls))
        # if len(urls) == len(current_urls):
        #     raise RuntimeError("No new data provided")

        logging.debug("Updating original data source")
        response = self.client.datasources.UpdateDataSource(
            json_format.ParseDict(
                {"id": self.id, "connection_info": connection_info},
                data_source.UpdateDataSourceRequest(),
            )
        )

        # if upload file from local, and it's not Excel file, we don't run pipeline
        job_id = response.job_id
        if job_id:
            wait_pipeline_run(self.client, job_id, timeout)

        data_source_id = self.id
        def data_source_processed():
            self.data = self.client.datasources.GetDataSource(
                data_source.GetDataSourceRequest(id=data_source_id)
            )
            if self.data.status == "data_ready":
                return True
            if self.data.status == STATUS_FAILED:
                raise RuntimeError(f"Failed to create DataSource(id={self.data.id})")

        wait_until(data_source_processed, timeout=timeout)

        old_data_items = [di.id for di in current_ds.data_items]
        # print("old data items: ", old_data_items)

        new_ds = self.client.datasources.GetDataSource(
            data_source.GetDataSourceRequest(id=self.id)
        )
        return [di.id for di in new_ds.data_items if di.id not in old_data_items]

class FileSource(DataSource):
    file_urls: List[str]  # Support local files, HTTP/HTTPS and S3
    file_type: FileType

    def __init__(
        self,
        file_urls: List[str],
        schema: List[Column] = [],
        file_type: str = FileType.CSV,
        file_settings: FileSettings = CSVSettings(),
        storage_options: dict = {},
    ) -> None:
        """A DataSource from files (local, HTTP/HTTPS, S3)

        Args:
            file_urls (List[str]): URLs of data files. Can be local file paths.
            schema (List[Column], optional): Schema of the data. Defaults to [].
            file_type (str, optional): Format of the data files. Can be CSV, JSONLine, Parquet or Excel. Defaults to FileType.CSV.
            file_settings (FileSettings, optional): _description_. Defaults to CSVSettings().
            storage_options (dict). Provide S3 credentials if required
                                    {
                                        'awsS3Endpoint': '',
                                        'awsRegion': '',
                                        'awsAccessKeyId': '',
                                        'awsSecretAccessKey': ''
                                    }
        """

        self.file_urls = file_urls
        self.file_type = file_type
        self.storage_options = storage_options

        if self.file_type in (FileType.Parquet, FileType.JSONLine):
            file_settings = FileSettings()

        settings_builder = SettingsBuilder(
            {"fileType": file_type, "settings": file_settings.asdict()}
        )
        settings_builder.set_file_type(self.file_type)
        self.file_settings = settings_builder.build()

        schema_builder = SchemaBuilder(schema)
        self.schema = schema_builder.build()

    def upload_files(self, project_id):
        # Step 1: Upload file if necessary
        file_type_mapping = {
            FileType.CSV: "text/csv",
            FileType.JSONLine: "application/jsonl",
            FileType.Parquet: "application/binary+parquet",
            FileType.Excel: "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
        }
        content_type = file_type_mapping[self.file_type]
        return [
            _upload_file(self.client, project_id, path, content_type)
            for path in self.file_urls
        ]

    def _create_connection_info(self, project_id: str):
        protocol = "s3"
        urls = None
        if self.file_urls[0].startswith('http://') or self.file_urls[0].startswith('https://'):
            protocol = "http"
            urls = self.file_urls
        elif self.file_urls[0].startswith('s3://'):
            urls = self.file_urls
        else:
            urls = self.upload_files(project_id)
        connection_info = {
            "urls": urls,
            "fileType": self.file_type,
        }
        if (self.storage_options is not None):
            connection_info.update(self.storage_options)

        return protocol, connection_info

class DatabaseType(str, Enum):
    MySQL = "mysql"
    PostgreSQL = "postgres"
    SQLServer = "sqlserver"
    MongoDB = "mongodb"

class DatabaseSource(DataSource):
    type: DatabaseType
    """Type of DBMS: mysql, sqlserver postgres, and mongodb"""
    host: str
    port: int
    username: str
    password: str
    database: str
    table: str = None
    query: str = None

    def __init__(
        self,
        type: str,
        host: str,
        port: int,
        username: str,
        password: str,
        database: str,
        table: str = None,
        query: str = None,
        schema: List[Column] = [],
    ) -> None:
        """A DataSource from databases (DBMS: MySQL, SQL Server, PostgreSQL, and MongoDB)

        Args:
            host (str): database server hostname
            port (int): database server port
            username (str): database username
            password (str): database password
            database (str): database name
            table (str, optional): database table name. Defaults to None.
            query (str, optional): query to select the data. Defaults to None.
            schema (List[Column], optional): Schema of the data. Defaults to [].
        """
        self.type = type
        self.host = host
        self.port = port
        self.username = username
        self.password = password
        self.database = database
        self.table = table
        self.query = query

        schema_builder = SchemaBuilder(schema)
        self.schema = schema_builder.build()

        self.file_settings = SettingsBuilder().set_file_type(FileType.Parquet).build()

    def _create_connection_info(self, project_id: str) -> Tuple[str, dict]:
        if self.table is None and self.query is None:
            raise Exception("Either `table` or `query` must be provided.")

        connection_info = {
            "type": self.type,
            "host": self.host,
            "port": self.port,
            "username": self.username,
            "password": self.password,
            "database": self.database,
            "table": self.table,
            "query": self.query,
        }

        return "database", connection_info

@dataclass
class DatasetSource(DataSource):
    dataset_id: str
    dataset_version: int = LATEST_VERSION

def _upload_file(
    client: Client, project_id: str, path: str, content_type="text/csv"
) -> str:
    """Upload a file to the project

    Args:
        client (Client): API Client
        project_id (_type_): ID of the project to which this file is saved
        path (_type_): path to local file
        content_type (str, optional): _description_. Defaults to 'text/csv'.

    Raises:
        Exception: _description_
    """
    file_name = os.path.basename(path)
    request = file.UploadFileRequest(
        name=file_name, content_type=content_type, project_id=project_id
    )

    response = client.files.UploadFile(request)
    file_url = response.url
    upload_url = response.upload_url

    # Upload data using the returned (pre-signed) URL
    with open(path, "rb") as data:
        upload_headers = {"Content-Type": content_type}
        response = requests.put(upload_url, data=data, headers=upload_headers)
        if response.status_code == 200:
            logging.debug(f"{path} is uploaded to {file_url}")
            return file_url
        else:
            raise Exception("Cannot upload data file: {}".format(response))

dtype_to_semantic_type = {
    'object': DataType.Text,
    'int': DataType.Numeric,
    'int16': DataType.Numeric,
    'int32': DataType.Numeric,
    'int64': DataType.Numeric,
    'int128': DataType.Numeric,
    'float': DataType.Numeric,
    'float32': DataType.Numeric,
    'float64': DataType.Numeric,
    'float128': DataType.Numeric,
    'datetime64[ns]': DataType.DateTime,
    'bool': DataType.Boolean,
}

def generate_schema(df):
    """Infer schema from the Pandas dataframe 

    Args:
        df (pandas.DataFrame): _description_

    Returns:
        List[Column]: inferred schema of the dataframe
    """
    return [Column(col, dtype_to_semantic_type[str(df.dtypes[col])]) for col in df.columns]

def print_schema(df):
    """Print schema of the Pandas DataFrame

    Args:
        df (pandas.DataFrame): Pandas DataFrame
    """
    schema = generate_schema(df)
    print("[")
    print(',\n'.join([f'    Column("{column.name}", {column.type})' for column in schema]))
    print("]")