from dataclasses import dataclass
import os
import json
from typing import List, NewType
import logging
from aiaengine.api import LATEST_VERSION
from aiaengine.api import file as api_file
from aiaengine.api import dataset as api_dataset
from aiaengine.api import data_source as api_data_source
from aiaengine.api import pipeline as api_pipeline
from aiaengine.common import DEFAULT_TIMEOUT, STATUS_FAILED, STATUS_SUCCESS
from aiaengine.data_source import DataSource
from aiaengine.common import Resource
from aiaengine.waiter import wait_until, wait_pipeline_run
from aiaengine.io_utils import download_files
from google.protobuf import json_format
from aiaengine.api import artifact as api_artifact

class DatasetAnalysisTask:
    def task_type(self) -> str:
        return None

    def config(self) -> dict:
        return {}

@dataclass
class DistributionAnalysisTask(DatasetAnalysisTask):
    grouping_column: str
    hist_nbins: int = None
    max_groups: int = None
    max_categories: int = None

    def task_type(self) -> str:
        return 'distribution_analysis'
    
    def config(self) -> dict:
        config = {
            'grouping_column': self.grouping_column
        }
        if self.hist_nbins is not None:
            config['hist_nbins'] = self.hist_nbins
        if self.max_groups is not None:
            config['max_groups'] = self.max_groups
        if self.max_categories is not None:
            config['max_categories'] = self.max_categories
        return config

@dataclass
class DatasetAnalysis:
    config: dict
    result: dict

# forward declaration
Dataset = NewType("Dataset", Resource)

class Dataset(Resource):
    """The object model represents a Dataset in the Engine"""

    @property
    def name(self) -> str:
        """Name of the dataset

        Returns:
            str: Name of the dataset
        """
        return self._data.name

    @property
    def description(self) -> str:
        """Description of the dataset

        Returns:
            str: Description of the dataset
        """
        return self._data.description

    def update(self, name: str, description: str = None):
        """Update dataset information

        Args:
            name (str): updated dataset name
            description (str, optional): updated dataset description. Defaults to None.
        """
        self.client.datasets.UpdateDataset(
            api_dataset.UpdateDatasetRequest(
                id=self.id,
                name=name, 
                description=description
            )
        )

    def delete(self):
        """Delete dataset
        """
        self.client.datasets.DeleteDataset(
            api_dataset.DeleteDatasetRequest(
                id=self.id
            )
        )

    def download(self, output_folder: str, version=LATEST_VERSION) -> List[str]:
        """Download dataset as parquet files into folder

        Args:
            output_folder (str): Folder to save data files

        Returns:
            List of downloaded file paths
        """
        list_files_response = self.client.files.ListFiles(
            api_file.ListFilesRequest(dataset_id=self.id, version=version)
        )
        urls_and_file_names = []
        files = list_files_response.files
        for f in files:
            r = self.client.files.GetFileDownloadUrl(
                api_file.GetFileDownloadUrlRequest(id=f.id, project_id=self._data.project.id)
            )
            urls_and_file_names.append((r.url, os.path.join(output_folder, f.name)))

        download_files(urls_and_file_names)

        return [file_name for _, file_name in urls_and_file_names]

    def copy(self, name: str, project_id: str = None, description: str='', version: int = -2) -> Dataset:
        """Copy an existing dataset to the project

        Args:
            project_id (str): ID of the project to which the dataset will be copied. Defaults to None (same project as current dataset)
            name (str): name of the copied dataset
            description (str, optional): description of the copied dataset. Defaults to ''.
            version (int, optional): version to copy: -2 = all versions, -1 = latest version, otherwise the exact version
            type (str, optional): dataset type, can be empty or 'intermediate'. Defaults to ''.

        Returns:
            the new dataset
        """
        if project_id is None:
            project_id = self._data.project.id

        new_dataset = self.client.datasets.CopyDataset(api_dataset.CopyDatasetRequest(
            id=self.id,
            project_id=project_id,
            name=name,
            description=description,
            version=version,
        ))
        return Dataset(id=new_dataset.id, data=new_dataset, client=self.client)
    
    def to_pandas(self, version: int = LATEST_VERSION):
        """Read dataset to Pandas DataFrame

        Args:
            version (int, optional): dataset version. Defaults to LATEST_VERSION.

        Returns:
            pd.DataFrame: the dataset in form of a Pandas DataFrame
        """
        import pandas as pd

        list_files_response = self.client.files.ListFiles(
            api_file.ListFilesRequest(dataset_id=self.id, version=version)
        )
        files = list_files_response.files
        dataframes = []
        for f in files:
            r = self.client.files.GetFileDownloadUrl(
                api_file.GetFileDownloadUrlRequest(id=f.id, project_id=self._data.project.id)
            )
            dataframes.append(pd.read_parquet(r.url))
        return pd.concat(dataframes)

    def update_data(self, data_source: DataSource, timeout: int = DEFAULT_TIMEOUT) -> int:
        """Update a dataset with new data

        Args:
            data_source (DataSource): Data source from which this dataset updated
            timeout (int, optional): time to wait for data to be updated. Defaults to DEFAULT_TIMEOUT.

        Raises:
            RuntimeError: runtime error

        Returns:
            int: latest version of the dataset
        """
        dataset_info = self.client.datasets.GetDataset(
            api_dataset.GetDatasetRequest(id=self.id)
        )
        if dataset_info.data_source is None:
            raise RuntimeError(
                "Could not update a dataset which was not originally imported from a data source"
            )
        data_source_id = dataset_info.data_source.id

        data_source.client = self.client
        data_source.id = data_source_id
        new_data_item_ids = data_source.update(timeout=timeout)
        if len(new_data_item_ids) == 0:
            raise RuntimeError("No new data items available for updating the dataset")

        response = self.client.datasources.CreateDatasetVersion(
            json_format.ParseDict(
                {
                    "id": data_source_id,
                    "dataset_id": self.id,
                    "data_item_ids": new_data_item_ids,
                },
                api_data_source.CreateDatasetVersionRequest(),
            )
        )

        def dataset_updated():
            rs = self.client.pipelines.GetResult(
                api_pipeline.GetResultRequest(job_id=response.job_id)
            )
            if rs.status == STATUS_SUCCESS:
                ds = self.client.datasets.GetDataset(
                    api_dataset.GetDatasetRequest(id=rs.results[0].result["datasetId"])
                )
                logging.info(
                    f"Dataset(id={ds.id}) has been updated. New version is {ds.latest_version}"
                )
                return ds
            elif rs.status == STATUS_FAILED:
                raise RuntimeError(f"Failed to update Dataset(id={self.id})")

        updated_dataset = wait_until(dataset_updated, timeout=timeout)
        self.latest_version = updated_dataset.latest_version
        return updated_dataset.latest_version

    def run_analysis(self, task: DatasetAnalysisTask, dataset_version: int = LATEST_VERSION, timeout: int = DEFAULT_TIMEOUT) -> DatasetAnalysis:
        step_parameters = [
            api_pipeline.StepParameter(
                step_name="advanced-data-analysis",
                parameters={
                    "PROJECT_ID": self._data.project.id,
                    "DATASET_ID": self.id,
                    "DATASET_VERSION": f"{dataset_version}",
                    "ANALYSIS_TASKS": json.dumps([{ 'task_type': task.task_type(), 'config': task.config() }])
                },
            ),
        ]

        pipeline_run = self.client.pipelines.CreatePipelineRun(
            api_pipeline.CreatePipelineRunRequest(
                pipeline_id="system.advanced-data-analysis",
                owner_type="dataset",
                owner_id=self.id,
                owner_metadata={
                    "actionName": "system.action.advanced-data-analysis"
                },
                step_parameters=step_parameters,
                global_parameters={
                    "ORG": self._data.org.id,
                    "PROJECT": self._data.project.id,
                },
            )
        )

        # wait for job completed
        pipeline_run = wait_pipeline_run(self.client, pipeline_run.id, timeout)        
        artifact_id = json.loads(pipeline_run.results[0].result['output_artifacts'])[0]
        artifact = self.client.artifacts.GetArtifact(api_artifact.GetArtifactRequest(id=artifact_id))
        data = json_format.MessageToDict(artifact.data)
        return DatasetAnalysis(
            config=data['config'],
            result=data['result']
        )

    def list_analyses(self) -> List[DatasetAnalysis]:
        """List all completed dataset analyses

        Returns:
            list of all completed dataset analysis results
        """
        response = self.client.artifacts.ListArtifacts(api_artifact.ListArtifactsRequest(
            type='data_analysis_result',
            labels={
                'project': self._data.project.id,
                'dataset': self.id,
            }
        ))
        results = []
        for artifact in response.artifacts:
            data = json_format.MessageToDict(artifact.data)
            results.append(DatasetAnalysis(
                config=data['config'],
                result=data['result']
            ))
        return results

    def _get_data(self) -> any:
        return self.client.datasets.GetDataset(
            api_dataset.GetDatasetRequest(id=self.id)
        )

    def __repr__(self) -> str:
        return f"Dataset(id='{self.id}', name='{self.name}', description='{self.description}')"
