from typing import Set, List
from aiaengine.api import featureset as api_feature_set
from aiaengine.common import DEFAULT_TIMEOUT
from aiaengine.common import Resource
from aiaengine.waiter import wait_featureset_processed
import pandas as pd
import json


class FeatureSet(Resource):
    """The object model represents a Feature Set in the Engine
    """

    @property
    def name(self) -> str:
        """Name of the feature set

        Returns:
            str: Name of the feature set
        """
        return self._get_data().name

    @property
    def description(self) -> str:
        """Description of the feature set

        Returns:
            str: Description of the feature set
        """
        return self._get_data().description

    @property
    def feature_names(self) -> Set[str]:
        """Selected column names that used as features

        Returns:
            Set[str]: set of column names that used as features
        """
        return set(self._get_data().feature_names)

    def select_recommended_models(self, n: int, by_metric: str) -> List[str]:
        """
            Get recommeneded models based on a given metric

        Args:
            n (int): top N recommended algorithms
            by_metric (str): ranking metric 

        Returns:
            List[str]: list of algorithm (model template name)
        """
        df_model_rank = pd.DataFrame(columns=["template", by_metric])
        model_recommendation = json.loads(self._data.recommendations)["recommendation"]
        for model in model_recommendation:
            if model["metrics"][by_metric] is not None:
                df_model_rank = df_model_rank.append(
                    {
                        "template": model["template_id"],
                        by_metric: model["metrics"][by_metric],
                    },
                    ignore_index=True,
                )
        selected_models = (
            df_model_rank.sort_values(by=by_metric, ascending=False)["template"]
            .head(n)
            .tolist()
        )
        return selected_models

    def wait_for_processed(self, timeout: int = DEFAULT_TIMEOUT):
        """Waiting for the feature set processed

        Args:
            timeout (int, optional): waiting time in seconds. Defaults to DEFAULT_TIMEOUT (600s).
        """
        # wait_for_processing(self.client, "featureset", response.id, timeout)
        feature_set_data = wait_featureset_processed(self.client, self.id, timeout=timeout)
        self._data = feature_set_data

    def _get_data(self) -> any:
        return self.client.featuresets.GetFeatureSet(
            api_feature_set.GetFeatureSetRequest(id=self.id)
        )

