""" IO Utils """
import json
import os
import tempfile
from io import BytesIO
from pathlib import Path
from typing import List, Tuple
from urllib.parse import urlparse
from urllib.request import urlopen
import boto3
import asyncio
import aiohttp
from contextlib import closing
CHUNK_SIZE = 4 * 1024  # 4 KB
from parfive import Downloader, SessionConfig

# pylint: disable=line-too-long, invalid-name
def get_s3():
    """
    Build s3 object using ENV CREDENTIAL
    """
    if "AWS_S3_ENDPOINT" in os.environ:
        endpoint = os.environ["AWS_S3_ENDPOINT"]
    elif "AWS_ENDPOINT" in os.environ:  # fallback
        endpoint = os.environ["AWS_ENDPOINT"]
    else:
        endpoint = "https://s3.amazonaws.com"
    if "AWS_REGION" in os.environ:
        region = os.environ["AWS_REGION"]
    else:
        region = "us-east-1"
    if not endpoint.startswith("http"):
        endpoint = "https://%s" % endpoint
    return boto3.resource(
        "s3",
        aws_access_key_id=os.environ["AWS_ACCESS_KEY_ID"],
        aws_secret_access_key=os.environ["AWS_SECRET_ACCESS_KEY"],
        endpoint_url=endpoint,
        region_name=region,
    )


def extract_bucket_and_key(s3_url):
    """
    Extract bucket and key from a s3 url
    """
    url_parts = urlparse(s3_url)

    return url_parts.netloc, url_parts.path.lstrip("/")


def is_s3_path(file_path):
    """
    Return True if file_path is a s3 url, otherwise False
    """
    return urlparse(file_path).scheme == "s3"


def is_s3_file_exists(s3_url):
    """
    Check if a s3 file is exist or not
    """
    if not is_s3_path(s3_url):
        raise Exception(f"{s3_url} is not a valid s3 url")

    s3 = get_s3()
    bucket_name, s3_key = extract_bucket_and_key(s3_url)
    bucket = s3.Bucket(bucket_name)
    s3_objs = list(bucket.objects.filter(Prefix=s3_key))
    return any([s3_obj.key == s3_key for s3_obj in s3_objs])


def list_urls(s3_dir, file_ext):
    """
    List all files in s3_dir with ext s3_dir
    """
    if not is_s3_path(s3_dir):
        raise Exception(f"{s3_dir} is not a valid s3 url")

    s3 = get_s3()
    bucket_name, s3_key = extract_bucket_and_key(s3_dir)
    bucket = s3.Bucket(bucket_name)
    s3_objs = list(bucket.objects.filter(Prefix=s3_key))

    return [
        f"s3://{bucket_name}/{str(obj.key)}"
        for obj in s3_objs
        if (str(obj.key)).endswith(file_ext)
    ]


def get_s3_size(path):
    """
    Get the size of the file
    :param path:
    :return: the size of specified path in bytes
    """
    path_elements = path.split("/")
    bucket = path_elements[2]

    file = "/".join(path_elements[3:])
    if (
        path.endswith("*.parquet")
        or path.endswith("*")
        or path.endswith("*.csv")
        or path.endswith("*.jsonl")
    ):
        file = "/".join(path_elements[3:-1])

    total_size = 0
    for obj in get_s3().Bucket(bucket).objects.filter(Prefix=file):
        file_name = str(obj.key)
        file_extension = (
            ".csv"
            if path.endswith("*.csv")
            else (
                ".parquet"
                if path.endswith("*.parquet")
                else ".jsonl"
                if path.endswith("*.jsonl")
                else None
            )
        )
        total_size += (
            obj.size
            if file_extension is None or file_name.endswith(file_extension)
            else 0
        )

    return total_size


def get_local_size(path):
    """
    Get file size of the path
    :param path: local file path
    :return: the size of specified path in bytes
    """
    if os.path.isfile(path):
        return os.path.getsize(path)

    total_size = 0
    for dirpath, dirnames, filenames in os.walk(
        path
    ):  # pylint: disable=unused-variable
        for f in filenames:
            fp = os.path.join(dirpath, f)
            total_size += os.path.getsize(fp)
    return total_size


def get_file_size(file_path):
    """
    Return file size
    """
    return (
        get_s3_size(file_path) if is_s3_path(file_path) else get_local_size(file_path)
    )


def is_file(file_path):
    """
    Check if the file_path (local or s3 url) is a file
    """
    return (
        is_s3_file_exists(file_path)
        if is_s3_path(file_path)
        else os.path.isfile(file_path)
    )


def download_data(url, path):
    """
    Download from s3
    """
    s3 = get_s3()
    bucket, key = extract_bucket_and_key(url)
    s3.Object(bucket, key).download_file(path)


def download_s3_folder(s3_url, local_dir=None):
    """
    Download the contents of a folder from s3 to a local folder
    Args:
        s3_url: s3 url
        local_dir: a relative or absolute directory path in the local file system
    """
    s3 = get_s3()
    bucket_name, s3_folder = extract_bucket_and_key(s3_url)
    bucket = s3.Bucket(bucket_name)
    for obj in bucket.objects.filter(Prefix=s3_folder):
        target = (
            obj.key
            if local_dir is None
            else os.path.join(local_dir, os.path.relpath(obj.key, s3_folder))
        )
        if not os.path.exists(os.path.dirname(target)):
            os.makedirs(os.path.dirname(target))
        if obj.key[-1] == "/":
            continue
        bucket.download_file(obj.key, target)


def delete_s3_folder(s3_url):
    """
    Delete all files under key s3_url
    """
    s3 = get_s3()
    bucket_name, s3_folder = extract_bucket_and_key(s3_url)
    bucket = s3.Bucket(bucket_name)
    bucket.objects.filter(Prefix=s3_folder).delete()


def upload_data(path, url):
    """
    Upload a file to s3
    """
    s3 = get_s3()
    bucket_name, key = extract_bucket_and_key(url)
    s3.Object(bucket_name, key).upload_file(path)


def create_folder(path: str):
    """
    Create a director
    """
    if is_s3_path(path):
        url = (path + "/") if not path.endswith("/") else path
        s3 = get_s3()
        bucket_name, key = extract_bucket_and_key(url)
        s3.Object(bucket_name, key + "/").put()
        print(f"created {url}")
    else:
        Path(path).mkdir(parents=True, exist_ok=True)


def download_data_to_memory(url):
    """
    Download data, load to memory
    """
    s3 = get_s3()
    bucket_name, key = extract_bucket_and_key(url)

    bytes_io = BytesIO()
    s3.Object(bucket_name, key).download_fileobj(bytes_io)
    return (bytes_io.getvalue()).decode("utf-8")


def upload_s3(data, url):
    """
    Put data to s3 url
    """
    (_, path) = tempfile.mkstemp()
    # save metadata as json
    with open(path, "w") as outfile:
        json.dump(data, outfile)
    # upload metadata file to the provided URL
    upload_data(path, url)


def upload_folder_to_s3(input_dir, s3_url):
    """
    Upload a local directory to s3
    """
    s3 = get_s3()
    bucket_name, s3_path = extract_bucket_and_key(s3_url)
    bucket = s3.Bucket(bucket_name)
    try:
        for path, _, files in os.walk(input_dir):
            for file in files:
                dest_path = path.replace(input_dir, "")
                __s3_file = os.path.normpath(s3_path + "/" + dest_path + "/" + file)
                __local_file = os.path.join(path, file)
                print("upload", __local_file, "to", __s3_file, end="")
                bucket.upload_file(__local_file, __s3_file)
                print(" ...success")
    except Exception as e:
        print("Upload failed!")
        print(e)
        raise e


def get_filename(url):
    """
    Return file name
    """
    return (urlparse(url)).path.split("/")[-1]


def load_file_to_json(file_url):
    """
    Return json content from a file url
    """
    result_json = None

    try:
        if file_url is not None:
            if file_url.startswith("http"):
                res = urlopen(file_url)
                data_bytes = res.read()
                result_json = json.loads(data_bytes.decode("utf-8"))
            else:
                if is_s3_path(file_url):
                    content_str = download_data_to_memory(file_url)
                    result_json = json.loads(content_str)
                else:
                    path_local = file_url
                    with open(path_local, "r") as f:
                        result_json = json.loads(f.read())

    except:
        print(f"error load file: {file_url}")
    return result_json


def get_storage_options(path):
    if not isinstance(path, str) or not path.startswith("s3://"):
        return None

    if os.environ["AWS_ENDPOINT"] != "":
        endpoint_url = os.environ["AWS_ENDPOINT"]
        if not endpoint_url.startswith("http"):
            endpoint_url = "https://" + endpoint_url
        return {
            "anon": False,
            "key": os.environ["AWS_ACCESS_KEY_ID"],
            "secret": os.environ["AWS_SECRET_ACCESS_KEY"],
            "client_kwargs": {"endpoint_url": endpoint_url},
        }
    return None


def load_json(url):
    """
    Load JSON content from a url
    """
    result_json = None

    try:
        if url is not None:
            if url.startswith("http"):
                res = urlopen(url)
                data_bytes = res.read()
                result_json = json.loads(data_bytes.decode("utf-8"))
            else:
                if is_s3_path(url):
                    content_str = download_data_to_memory(url)
                    result_json = json.loads(content_str)
                else:
                    path_local = url
                    with open(path_local, "r") as f:
                        result_json = json.loads(f.read())

    except:
        print(f"error load file: {url}")
    return result_json


def save_json(data, url):
    """Save JSON content to a url

    Args:
        json_data ([type]): [description]
        file_url ([type]): [description]
    """
    if is_s3_path(url):
        upload_s3(data, url)
    else:
        with open(url, "w") as f:
            json.dump(data, f)


def download_files(urls_and_file_names: List[Tuple[str, str]]):
    dl = Downloader(progress=True, overwrite=True, config=SessionConfig(file_progress=False))
    for (url, file_name) in urls_and_file_names:
        dl.enqueue_file(url, filename=file_name)
    files = dl.download()