import base64
from dataclasses import dataclass
from enum import Enum
import json
import logging
from aiaengine.api import LATEST_VERSION
from aiaengine.api.client import Client
from aiaengine.api import dataset as api_dataset
from aiaengine.api import model as api_model
from aiaengine.api import pipeline as api_pipeline
from aiaengine.api import app as api_app
from aiaengine.api import featureset as api_featureset
from aiaengine.api import artifact as api_artifact
from aiaengine.common import DEFAULT_TIMEOUT, PIPELINE_RUN_STATUS_SUCCESS, STATUS_SUCCESS, STATUS_FAILED, STATUS_ACTIVE
from aiaengine.data_source import DataSource, FileSource, DatabaseSource, DatasetSource
from aiaengine.dataset import Dataset
from aiaengine.common import Resource
from aiaengine.waiter import wait_until, wait_model_trained, wait_evaluation_processed, get_training, wait_pipeline_run
from aiaengine.dataframe_utils import load_dataframe
from google.protobuf import json_format
import requests

@dataclass
class EvaluationResult:
    """Result of model evaluation
    """
    summary: dict
    """Evaluation summary
    """

    details: dict
    """Evaluation result details
    """

class Evaluation(Resource):
    """The object model represents a model evaluation request and result in the Engine
    """
    result: EvaluationResult

    def __init__(self, id: str, data: any = None, client: Client = None) -> None:
        super().__init__(id, data, client)
        self.model_id = data.model.id

    def wait_for_processed(self, timeout: int = DEFAULT_TIMEOUT):
        evaluation_data = wait_evaluation_processed(self.client, model_id=self.model_id, evaluation_id=self.id, timeout=timeout)

        result = json_format.MessageToDict(evaluation_data.result)
        self.result = EvaluationResult(summary=result["summary"], details=result["details"])
        self._data = evaluation_data

class Prediction(Resource):
    """The object model represents a prediction request in the Engine
    """
    result: Dataset = None

    def __init__(self, id: str, data=None, client: Client = None) -> None:
        super().__init__(id, data, client)
    
    def _get_data(self) -> any:
        return self.client.models.GetPrediction(api_model.GetPredictionRequest(id=self.id))
    
    def delete(self):
        return self.client.models.DeletePrediction(api_model.DeletePredictionRequest(
            id=self._data.model.id,
            prediction_id=self.id))

    def wait_for_processed(self, timeout: int = DEFAULT_TIMEOUT):
        def prediction_processed():
            prediction = self.client.models.GetPrediction(
                api_model.GetPredictionRequest(id=self.id)
            )
            if prediction.status == STATUS_SUCCESS:
                return prediction
            elif prediction.status == STATUS_FAILED:
                raise Exception(f"Failed to process the Prediction(id={self.id})")

        prediction = wait_until(prediction_processed, timeout=timeout)

        self.result = Dataset(id=prediction.output_dataset.id, client=self.client)

    def export(self, dataset_name: str, project_id: str=None, timeout: int = DEFAULT_TIMEOUT):
        """
            Export prediction result dataset to the project.

            Args:
            dataset_name (str): name of the copied dataset
            project_id (str): ID of the project to which the dataset will be copied. Defaults to None (same project as current dataset)
        """
        export_project = self._data.project.id if project_id is None else project_id
        # Get org from project
        from aiaengine.api.project_pb2 import GetProjectRequest
        project = self.client.projects.GetProject(GetProjectRequest(id=self._data.project.id))

        response = self.client.pipelines.CreatePipelineRun(
            api_pipeline.CreatePipelineRunRequest(
                pipeline_id="system.export-batch-prediction-output",
                owner_type="prediction",
                owner_id=self.id,
                owner_metadata={
                    "actionName": "system.action.export-batch-prediction-output",
                },
                step_parameters=[
                    api_pipeline.StepParameter(
                        step_name="export-prediction-output",
                        parameters={
                            "PREDICTION_ID": self.id,
                            "EXPORT_INFO": json.dumps({
                                "datasets":[{"projectId": export_project, "datasetName": dataset_name}]
                            })
                        },
                    )
                ],
                global_parameters={
                    "ORG": project.org.id,
                    "PROJECT": self._data.project.id,
                    "APP": self._data.app.id,
                    "MODEL": self._data.model.id
                },
            )
        )
        # wait for job completed
        pipeline_run = wait_pipeline_run(self.client, response.id, timeout)
        exported_info = json.loads(pipeline_run.results[0].result['exportedInfo'])
        exported_ds = [ds for ds in exported_info['exportedDatasets'] if ds['name'] == dataset_name][0]
        print(f"Exported dataset={exported_ds['_id']}")
        return Dataset(id=exported_ds['_id'], client=self.client)

class Endpoint(Resource):
    def __init__(self, id: str, app_id: str, client: Client = None) -> None:
        self.app_id = app_id
        super().__init__(id, None, client)
        self.url = self._data.url
        self.api_keys = self._data.api_keys

    def wait_for_ready(self, timeout: int = DEFAULT_TIMEOUT):
        def ping_endpoint(url: str) -> bool:
            try:
                res = requests.get(url=url)
                return res.status_code == 200
            except:
                return False

        def endpoint_ready():
            if ping_endpoint('%s/ping' % self.url):
                return self

        wait_until(endpoint_ready, timeout=timeout)
        return self

    def _get_data(self) -> any:
        return self.client.apps.GetEndpoint(api_app.GetEndpointRequest(
                id=self.app_id,
                endpoint_id=self.id
            ))

class Deployment(Resource):
    """The object model represents a deployment request in the Engine
    """

    app_id: str = None

    def __init__(self, id: str, app_id: str, client: Client = None) -> None:
        self.app_id = app_id
        super().__init__(id, None, client)

    def delete(self):
        """Delete the deployment
        """
        self.client.apps.DeleteDeployment(api_app.DeleteDeploymentRequest(
            id=self.app_id,
            deployment_id=self.id
        ))

    def predict(self, data_source: DataSource, explain: bool = False) -> any:
        """Run online prediction

        Args:
            data_source (DataSource): input data to run the prediction
            explain (bool, optional): set to True of the prediction explanation is expected. Defaults to False.

        Returns:
            Prediction result and explanations if explain = True
        """
        if not isinstance(data_source, FileSource):
            raise Exception("Only FileSource is supported at the moment")
        if len(data_source.file_urls) == 0:
            raise Exception("No input file")

        df = load_dataframe(data_source.file_urls[0])
        deployment = self._get_data()
        model = self.client.models.GetModel(api_model.GetModelRequest(id=deployment.model.id))
        featureset = self.client.featuresets.GetFeatureSet(api_featureset.GetFeatureSetRequest(id=model.feature_set.id))
        columns = list(featureset.feature_names)
        df_for_prediction = df[columns]
        endpoint = self.get_endpoint()
        if endpoint is None:
            raise Exception("Endpoint is not ready")

        api_key = ''
        for key in endpoint.api_keys:
            api_key = key.key
            break

        api_to_call = "explain" if explain else "invocations"
        result = self.call_endpoint(columns, df_for_prediction.values, f'{endpoint.url}/{api_to_call}?prediction_col_name=prediction'
                                                                      '&probabilities_score_col_name=probabilities', api_key)
        return result

    def call_endpoint(self, columns, lines, url, api_key) -> any:
        data = ','.join(str(x) for x in columns) + '\n'
        for index, line in enumerate(lines):
            if index == len(lines) - 1:
                data += ','.join(str(x) for x in line)
            else:
                data += ','.join(str(x) for x in line) + '\n'
        headers = {'Content-Type': 'text/csv', 'X-Api-Key': api_key}
        res = requests.post(url, data=data.encode('utf-8'), headers=headers)
        if res.status_code == 200:
            return res.text
        raise ValueError(res.text)

    def _get_data(self) -> any:
        return self.client.apps.GetDeployment(api_app.GetDeploymentRequest(id=self.app_id, deployment_id=self.id))

    def wait_for_processed(self, timeout: int = DEFAULT_TIMEOUT):
        def deployment_processed():
            response = self._get_data()
            if response.status == STATUS_ACTIVE:
                return response
            elif response.status == STATUS_FAILED:
                raise RuntimeError(f"Failed to process Deployment(id={id})")

        deployment = wait_until(deployment_processed, timeout=timeout)
        return deployment

    def get_endpoint(self, timeout: int = DEFAULT_TIMEOUT):
        endpoint = Endpoint(id=self._data.endpoint.id, app_id=self.app_id)
        return endpoint.wait_for_ready(timeout=timeout)

class ModelInsightType(str, Enum):
    Summary = "model_insights_summary"
    FeatureImportance = "feature_importance"
    ExplainedPredictions = "explained_predictions"

@dataclass
class ModelInsight:
    type: ModelInsightType
    data: dict

class Model(Resource):
    """The object model represents a project in the Engine
    """

    @property
    def name(self) -> str:
        """Name of the model

        Returns:
            Name of the model
        """
        return self._data.name

    def update(self, name: str):
        """Update model information

        Args:
            name (str): updated model name
            description (str, optional): updated model description. Defaults to None.
        """
        self.client.models.UpdateModel(
            api_model.UpdateModelRequest(
                id=self.id,
                name=name,
            )
        )

    def enable_continuous_learning(self, enabled: bool):
        """Enable/disable continuous learning for the model

        Args:
            enabled (bool): enabled if True
        """
        self.client.models.UpdateModel(
            api_model.UpdateModelRequest(
                id=self.id,
                name=self._data.name,
                continuous_learning=enabled,
            )
        )

    def delete(self):
        """Delete the model.

            Warning:
                This action cannot be undone. All trained model data will be deleted.
        """
        self.client.models.DeleteModel(
            api_model.DeleteModelRequest(
                id=self.id
            )
        )

    def wait_for_trained(
        self, dataset_version: int = LATEST_VERSION, timeout: int = DEFAULT_TIMEOUT * 2
    ):
        """Wait until the model trained

        Args:
            dataset_version (int, optional): the version of dataset on which the model was trained. Defaults to LATEST_VERSION.
            timeout (int, optional): time to wait for the model trained. Defaults to DEFAULT_TIMEOUT*2.
        """
        wait_model_trained(self.client, self.id, dataset_version, timeout)

    def evaluate(
        self,
        train_dataset_version: int = LATEST_VERSION,
        test_dataset_id: str = None,
        test_dataset_version: int = LATEST_VERSION,
        timeout: int = DEFAULT_TIMEOUT,
    ) -> Evaluation:
        """Evaluate the model with a test dataset

        Args:
            train_dataset_version (int, optional): the dataset version on which the model is trained. Defaults to LATEST_VERSION.
            test_dataset_version (int, optional): the version of test dataset. Defaults to LATEST_VERSION.
            timeout (int, optional): time to wait for the evaluation processed. Defaults to DEFAULT_TIMEOUT.

        Raises:
            RuntimeError: if the model has not been evaluated on the provided version of test dataset

        Returns:
            the Evaluation object
        """

        model_data = self.client.models.GetModel(api_model.GetModelRequest(id=self.id))
        training = self.get_training(dataset_version=train_dataset_version)
        if test_dataset_id is None:
            print(f'get test datast id from app')
            app = self.client.apps.GetApp(api_app.GetAppRequest(id=model_data.app.id))
            test_dataset_id = app.test_dataset_id

        response = self.client.models.ListEvaluations(
            api_model.ListEvaluationsRequest(
                app_id=model_data.app.id, model_id=self.id, training_id=training.id,
            )
        )
        latest_version = LATEST_VERSION
        evaluation_data = None
        for evaluation_data in response.evaluations:
            print(f'evaluation_data.test_dataset.id={evaluation_data.test_dataset.id}, version={evaluation_data.test_dataset.version}')
            if test_dataset_id == evaluation_data.test_dataset.id and \
                test_dataset_version == evaluation_data.test_dataset.version:
                evaluation_data = evaluation_data
                print('found')
                break
            if test_dataset_id == evaluation_data.test_dataset.id and \
                latest_version < evaluation_data.test_dataset.version:
                latest_version = evaluation_data.test_dataset.version
                print('try next')
                evaluation_data = evaluation_data

        evaluation = None
        if evaluation_data is None:
            print(f'create evaluation for test datast id {test_dataset_id}')
            # run evaluation
            evaluation_data = self.client.models.CreateEvaluation(api_model.CreateEvaluationRequest(
                id=self.id,
                training_id=training.id,
                test_dataset=api_dataset.RefDataset(
                    id=test_dataset_id,
                    version=test_dataset_version
                )
            ))

        evaluation = Evaluation(
            id=evaluation_data.id,
            data=evaluation_data,
            client=self.client
        )
        evaluation.wait_for_processed(timeout)

        return evaluation

    def _run_batch_prediction_on_dataset(self,
                                      dataset_id: str,
                                      dataset_version: int = LATEST_VERSION,
                                      train_dataset_version: int = LATEST_VERSION,
                                      timeout=DEFAULT_TIMEOUT,
                                      pipeline_configuration_id: str = '',
                                      ) -> Prediction:
        """Run a batch prediction

        Args:
            dataset_id (str): ID of the dataset to run the prediction
            dataset_version: version of dataset. Defaults to LATEST_VERSION.
            train_dataset_version (int, optional): Version of the train dataset on which the model trained. Defaults to LATEST_VERSION.
            timeout (int, optional): Wait maximum `timeout` seconds. Defaults to DEFAULT_TIMEOUT.

        Raises:
            RuntimeError: _description_

        Returns:
            Prediction: Prediction dataset
        """
        training = self.get_training(dataset_version=train_dataset_version)
        if training is None:
            raise RuntimeError(
                f"Model(id={self.id}) was not been trained on dataset version {train_dataset_version}"
            )
        step_parameters = [
            api_pipeline.StepParameter(
                step_name="create-prediction",
                parameters={
                    "DATASET_ID": dataset_id,
                    "DATASET_VERSION": str(dataset_version),
                    "MODEL_ID": self.id,
                    "TRAINING_ID": training.id,
                    "IS_INCLUDE_VARIABLE": "true",
                    "PIPELINE_CONFIGURATION_ID": pipeline_configuration_id,
                },
            ),
        ]
        prediction = self.run_prediction_pipeline(step_parameters=step_parameters,
                                                  enable_data_processing="false",
                                                  timeout=timeout)
        return prediction

    def run_batch_prediction(
        self,
        data_source: DataSource,
        train_dataset_version: int = LATEST_VERSION,
        timeout=DEFAULT_TIMEOUT,
        pipeline_configuration_id: str = ''
    ) -> Prediction:
        """Run a batch prediction

        Args:
            data_source (DataSource): source of the input data
            train_dataset_version (int, optional): Version of the train dataset on which the model trained. Defaults to LATEST_VERSION.
            timeout (int, optional): Wait maximum `timeout` seconds. Defaults to DEFAULT_TIMEOUT.
            pipeline_configuration_id: configuration used to run the prediction
        Raises:
            RuntimeError: _description_

        Returns:
            Prediction: Prediction resource
        """
        if not isinstance(data_source, FileSource) and not isinstance(data_source, DatasetSource):
            raise Exception("Only FileSource or DatasetSource are supported at the moment")

        if isinstance(data_source, DatasetSource):
            return self._run_batch_prediction_on_dataset(
                dataset_id=data_source.dataset_id,
                dataset_version=data_source.dataset_version,
                train_dataset_version=train_dataset_version,
                timeout=timeout,
                pipeline_configuration_id=pipeline_configuration_id,
            )

        model_data = self.client.models.GetModel(api_model.GetModelRequest(id=self.id))
        training = self.get_training(dataset_version=train_dataset_version)
        if training is None:
            raise RuntimeError(
                f"Model(id={self.id}) was not been trained on dataset version {train_dataset_version}"
            )

        logging.info("Uploading prediction data")
        data_source.client = self.client
        uploaded_file_urls = data_source.upload_files(model_data.project.id)

        params = base64.b64encode(
            json.dumps(
                {
                    "urls": uploaded_file_urls,
                    "fileType": data_source.file_type,
                    "nanValue": "N/A",
                }
            ).encode("utf-8")
        )

        step_parameters = [
            api_pipeline.StepParameter(
                step_name="data-processing",
                parameters={
                    "PARAMS": params,
                    "MODEL_ID": self.id,
                    "PROJECT_ID": model_data.project.id,
                },
            ),
            api_pipeline.StepParameter(
                step_name="create-prediction",
                parameters={
                    "MODEL_ID": self.id,
                    "TRAINING_ID": training.id,
                    "IS_INCLUDE_VARIABLE": "true",
                    "PIPELINE_CONFIGURATION_ID": pipeline_configuration_id,
                },
            ),
        ]
        prediction = self.run_prediction_pipeline(step_parameters=step_parameters,
                                                  enable_data_processing="true",
                                                  timeout=timeout)
        return prediction

    def run_prediction_pipeline(self,
                                step_parameters,
                                enable_data_processing: str,
                                timeout=DEFAULT_TIMEOUT,
                                pipeline_configuration_id: str = ''):
        model_data = self.client.models.GetModel(api_model.GetModelRequest(id=self.id))
        response = self.client.pipelines.CreatePipelineRun(
            api_pipeline.CreatePipelineRunRequest(
                pipeline_id="system.model-batch-prediction-preprocessing",
                owner_type="model",
                owner_id=self.id,
                owner_metadata={
                    "actionName": "system.action.model-batch-prediction-preprocessing"
                },
                step_parameters=step_parameters,
                global_parameters={
                    "ORG": model_data.org.id,
                    "PROJECT": model_data.project.id,
                    "MODEL": self.id,
                    "ENABLE-DATA-PROCESSING": enable_data_processing,
                },
                pipeline_configuration_id=pipeline_configuration_id,
            )
        )

        # wait for job completed
        pipeline_run = wait_pipeline_run(self.client, response.id, timeout=timeout)
        create_prediction_result = list(
            filter(lambda s: s.name == "default", pipeline_run.results)
        )[0]
        prediction_id = create_prediction_result.result["predictionId"]
        prediction = Prediction(id=prediction_id, client=self.client)
        prediction.wait_for_processed()

        return prediction

    def list_predictions(self, input_dataset_id:str=None):
        response = self.client.models.ListPredictions(api_model.ListPredictionsRequest(id=self.id))
        if input_dataset_id is None:
            return [Prediction(id=p.id, data=p, client=self.client) for p in response.predictions]
        else:
            return [Prediction(id=p.id, data=p, client=self.client) for p in response.predictions
                    if p.input_dataset.id == input_dataset_id]

    def get_training(self, dataset_version: int = LATEST_VERSION):
        """Get the training on dataset version

        Args:
            dataset_version (int, optional): version of the training dataset. Defaults to LATEST_VERSION.
        """
        return get_training(self.client, self.id, dataset_version)

    def deploy(
            self,
            name: str = None,
            description: str = None,
            endpoint_name: str = None,
            endpoint_id: str = None,
            endpoint_description: str = None,
            training_id: str = None,
            type: str = 'cloud',
            auto_activate: bool = True,
            auto_deploy: bool = False,
            timeout=DEFAULT_TIMEOUT,
    ) -> Deployment:
        """Deploy the model

        Args:
            name (str, optional): name of the deployment. Defaults to None.
            description (str, optional): description of the deployment. Defaults to None.
            endpoint_name (str, optional): name of the endpoint. Defaults to None.
            endpoint_id (str, optional): ID of the endpoint. Defaults to None.
            endpoint_description (str, optional): description of the endpoint. Defaults to None.
            training_id (str, optional): ID of the training. Defaults to None.
            type (str, optional): _description_. Defaults to 'cloud'.
            auto_activate (bool, optional): _description_. Defaults to True.
            auto_deploy (bool, optional): _description_. Defaults to False.
            timeout (_type_, optional): _description_. Defaults to DEFAULT_TIMEOUT.

        Returns:
            Deployment: the object represents the deployment
        """

        model = self._get_data()
        if training_id is None:
            training_id = model.last_success_training.id

        if not name:
            name = model.name

        if not endpoint_name:
            endpoint_name = 'Endpoint of %s deployment' % name

        target_info = api_model.TargetInfo(name=endpoint_name, id=endpoint_id, type="endpoint", description=endpoint_description)
        response = self.client.models.DeployModel(api_model.DeployModelRequest(
            auto_activate=auto_activate,
            auto_deploy=auto_deploy,
            id=self.id,
            training_id=training_id,
            target_info=target_info,
            type=type,
            name=name,
            description=description
        ))
        deployment = Deployment(id=response.id, app_id=model.app.id, client=self.client)
        deployment.wait_for_processed(timeout=timeout)
        return deployment

    def get_prediction_of_test_portion(
        self, 
        train_dataset_version: int = LATEST_VERSION,
        test_dataset_version: int = LATEST_VERSION,
        )-> Dataset:
        evaluation = self.evaluate(train_dataset_version, test_dataset_version)
        print(f"Evaluation id={evaluation.id}")
        response = self.client.pipelines.ListPipelineRuns(
            api_pipeline.ListPipelineRunsRequest(
                pipeline_id='system.model-evaluation',
                owner_type='model',
                owner_id=self.id,
                metadata={'evaluationId': evaluation.id}
            ))
        dataset_id = response.pipelineruns[0].results[0].result['outputDatasetId']
        print(f"Prediction dataset id={dataset_id}")
        return Dataset(dataset_id, client=self.client)

    def get_insight(self, type: ModelInsightType, dataset_version: int = LATEST_VERSION) -> ModelInsight:
        """Get insights 

        Args:
            type (ModelInsightsType): _description_
            dataset_version (int, optional): _description_. Defaults to LATEST_VERSION.

        Returns:
            ModelInsights: _description_
        """
        training = self.get_training(dataset_version=dataset_version)
        response = self.client.artifacts.ListArtifacts(api_artifact.ListArtifactsRequest(
            type=type,
            labels={
                'project': self._data.project.id,
                'model': self.id,
                'training': training.id
            }
        ))
        if len(response.artifacts) == 0:
            raise RuntimeError(f"Model insight {type} not found. Use generate_insights() to generate the insight first.")
        data = json_format.MessageToDict(response.artifacts[0].data)
        return ModelInsight(
            type=type,
            data=data
        )

    def generate_insights(self, dataset_version: int = LATEST_VERSION, timeout: int = DEFAULT_TIMEOUT):
        """Generate model insights

        Args:
            dataset_version (int, optional): the dataset version on which the model trained. Defaults to LATEST_VERSION.
            timeout (int, optional): time (in seconds) to wait for the model insights generated. Defaults to DEFAULT_TIMEOUT.

        Raises:
            RuntimeError: raised if the pipeline was failed

        Return:
            List of all generated artifacts
        """
        training = self.get_training(dataset_version=dataset_version)

        # check if already generated
        response = self.client.pipelines.ListPipelineRuns(api_pipeline.ListPipelineRunsRequest(
            pipeline_id="system.model-insights",
            owner_type="model",
            owner_id=self.id
        ))

        # only run if there is no successful run
        if len([p for p in response.pipelineruns if p.status == PIPELINE_RUN_STATUS_SUCCESS]) == 0:
            step_parameters = [
                api_pipeline.StepParameter(
                    step_name="model-insights",
                    parameters={
                        "MODEL_ID": self.id,
                        "TRAINING_ID": training.id,
                    },
                ),
            ]

            pipeline_run = self.client.pipelines.CreatePipelineRun(
                api_pipeline.CreatePipelineRunRequest(
                    pipeline_id="system.model-insights",
                    owner_type="model",
                    owner_id=self.id,
                    owner_metadata={
                        "actionName": "system.action.model_insights"
                    },
                    step_parameters=step_parameters,
                    global_parameters={
                        "ORG": self._data.org.id,
                        "PROJECT": self._data.project.id,
                        "MODEL": self.id,
                    },
                )
            )

            # wait for job completed
            wait_pipeline_run(self.client, pipeline_run.id, timeout)

        response = self.client.artifacts.ListArtifacts(api_artifact.ListArtifactsRequest(
            labels={
                'project': self._data.project.id,
                'model': self.id,
                'training': training.id
            }
        ))
        model_insights = []
        for artifact in response.artifacts:
            data = json_format.MessageToDict(artifact.data)
            model_insights.append(ModelInsight(
                type=artifact.type,
                data=data
            ))
        return model_insights

    def _get_data(self) -> any:
        return self.client.models.GetModel(api_model.GetModelRequest(id=self.id))

    def __repr__(self) -> str:
        return f"Model(id='{self.id}', name='{self.name}')"
