from dataclasses import dataclass
from typing import List
import random
import string
from aiaengine.project import Project
from aiaengine.common import Resource
from aiaengine.user import User, UserRoles
from aiaengine.api import org as api_org
from aiaengine.api import project as api_project
from aiaengine.api import user as api_user
from aiaengine.common import get_default_client
from aiaengine.util import unpack_grpc_error
from grpc import RpcError

@dataclass
class OrgUser:
    """The object model represents a user in the organization
    """

    user: User
    """User information"""
    role: str
    """Role of user in the organization"""
    external: bool = False
    """True if this user is an external user (invited to collaborate in the projects of the org)"""

class Org(Resource):
    """The object model represents an org in the Engine
    """

    @property
    def name(self) -> str:
        """Name of the organization

        Returns:
            Name of the organization
        """
        return self._data.name

    @property
    def description(self) -> str:
        """Description of the organization

        Returns:
            Description of the organization
        """
        return self._data.description

    def update(self, name: str, description: str = None):
        """Update information of the organization

        Args:
            name (str): new name of the organization
            description (str, optional): description of the organization. Defaults to None.
        """
        self.client.orgs.UpdateOrg(
            api_org.UpdateOrgRequest(
                id=self.id, name=name, description=description, logo_url=None
            )
        )

    def delete(self):
        """Delete the organization

            Warning:
                * Only organization which no active subscriptions can be deleted.
                * This action cannot be undone. All related data will be deleted.
        """
        self.client.orgs.DeleteOrg(api_org.DeleteOrgRequest(id=self.id))

    def create_project(self, name: str, description: str = "") -> Project:
        """Create a new project in the org

        Args:
            name (str): Name of the project
            description (str, optional): Description of the project. Defaults to ''.

        Returns:
            The newly created Project resource
        """
        project_data = self.client.projects.CreateProject(
            api_project.CreateProjectRequest(
                org_id=self.id, name=name, description=description
            )
        )
        return Project(id=project_data.id, data=project_data, client=self.client)

    def get_project(self, id: str) -> Project:
        """Get Project resource by ID

        Args:
            id (str): ID of the project

        Returns:
            Project: the requested Project resource
        """
        return Project(id=id, client=self.client)

    def list_projects(self) -> List[Project]:
        """List all projects of the organization

        Returns:
            list of `Project` objects
        """
        response = self.client.projects.ListOrgProjects(
            api_project.ListOrgProjectsRequest(org_id=self.id)
        )
        return [
            Project(id=project_data.id, data=project_data, client=self.client)
            for project_data in response.projects
        ]

    def create_user(
        self,
        name: str,
        email: str,
        password: str = None,
        role: str = UserRoles.OrgViewer,
    ) -> OrgUser:
        """Create a new organization user

        Args:
            name (str): Name of the user
            email (str): User email address
            password (str, optional): If password is None, password will be autogenrated
        """
        if password is None:
            # generate random password
            characters = string.ascii_letters + string.digits
            password = "".join(random.choice(characters) for i in range(8))
        user = self.client.users.CreateUser(
            api_user.CreateUserRequest(name=name, email=email, password=password)
        )
        # add user to org
        self.client.orgs.CreateOrgUser(
            api_org.CreateOrgUserRequest(id=self.id, user_id=user.id, role_id=role)
        )

        return OrgUser(
            user=User(id=user.id, name=user.display_name, email=email),
            role=role
        )

    def add_user(self, user_email: str, role: str = UserRoles.OrgViewer):
        """Add an existing user to the organization

        Args:
            user_id (str): ID of the user
            role (str, optional): Role of the user in the org. Defaults to UserRoles.OrgViewer.
        """
        try:
            user = self.client.users.GetUserByEmail(api_user.GetUserByEmailRequest(
                email=user_email
            ))
            self.client.orgs.CreateOrgUser(
                api_org.CreateOrgUserRequest(id=self.id, user_id=user.id, role_id=role,)
            )
        except RpcError as e:
            error_message, error_code = unpack_grpc_error(e)
            raise RuntimeError(f"{error_code}: {error_message}")

    def remove_user(self, user_email: str):
        """Remove a user from the organization

        Args:
            user_id (str): Unique identity of the user
        """
        try:
            user = self.client.users.GetUserByEmail(api_user.GetUserByEmailRequest(
                email=user_email
            ))

            self.client.orgs.DeleteOrgUser(
                api_org.DeleteOrgUserRequest(id=self.id, user_id=user.id)
            )
        except RpcError as e:
            error_message, error_code = unpack_grpc_error(e)
            raise RuntimeError(f"{error_code}: {error_message}")

    def list_users(self):
        """Return all users of the organization
        """
        response = self.client.orgs.ListOrgUsers(
            api_org.ListOrgUsersRequest(id=self.id)
        )

        return [
            OrgUser(
                user=User(
                    id=org_user.user.id,
                    name=org_user.user.display_name,
                    email=org_user.user.email
                ), 
                role=org_user.role
            )
            for org_user in response.org_users
        ] + [
            OrgUser(
                user=User(
                    id=org_user.user.id,
                    name=org_user.user.display_name,
                    email=org_user.user.email
                ), 
                role=org_user.role,
                external=True
            )
            for org_user in response.external_users
        ]

    def list_user_orgs():
        """List all organizations of the current authenticated user
        """
        client = get_default_client()
        response = client.orgs.ListOrgs(api_org.ListOrgsRequest(user_id=client.user_id))
        return [
            Org(id=org.id, data=org, client=client)
            for org in response.orgs
        ]

    def _get_data(self) -> any:
        return self.client.orgs.GetOrg(api_org.GetOrgRequest(id=self.id))

    def __repr__(self) -> str:
        return f"Org(id='{self.id}', name='{self.name}', description='{self.description}')"