from abc import abstractmethod
import os
import json
from inspect import signature, Parameter, _empty
import logging
import traceback
import typing
import grpc
from aiaengine.util import stop_on_failed, unpack_grpc_error
from aiaengine.errors import ValidationError

class Converter:
    """Convert a serialized input value in string format to Python object"""
    @abstractmethod
    def convert_from(self, value: str) -> any:
        pass

    @abstractmethod
    def convert_to(self, value: any) -> str:
        pass

class ServiceProvider:
    @abstractmethod
    def get_instance(self) -> any:
        pass

_SETTINGS = {
    'enable_task_decoration': False
}
_CONVERTERS = {}
_SERVICE_PROVIDERS = {}

TASK_INPUT_PREFIX = 'AIA_ENGINE_PIPELINE_TASK_INPUT_'

def enable_task_decoration(enabled=True):
    _SETTINGS['enable_task_decoration'] = enabled

def register_converter(type_name: str, converter: Converter):
    _CONVERTERS[type_name] = converter

def get_converter(type_name: str) -> Converter:
    if type_name not in _CONVERTERS:
        raise Exception(f'Could not find converter for type: {type_name}')
    return _CONVERTERS[type_name]

def register_service_provider(type_name: str, provider: ServiceProvider):
    _SERVICE_PROVIDERS[type_name] = provider

def get_service_provider(type_name: str) -> ServiceProvider:
    if type_name not in _SERVICE_PROVIDERS:
        raise Exception(f'Could not find service provider for type: {type_name}')
    return _SERVICE_PROVIDERS[type_name]

def set_inputs(args: dict):
    for name, value in args.items():
        os.environ[f'{TASK_INPUT_PREFIX}{name.upper()}'] = str(value)

def format_object_input(object_type: str, object_id: str, **attribues):
    data = { 'type': object_type, 'id': object_id }
    data.update(**attribues)
    return json.dumps(data)

def set_object_input(input_name: str, object_type: str, object_id: str, **attribues):
    set_inputs({
        input_name: format_object_input(object_type=object_type, object_id=object_id, **attribues)
    })

def _get_argument(p: Parameter):
    value = None if p.default is _empty else p.default
    env_var = f'{TASK_INPUT_PREFIX}{p.name.upper()}'
    if env_var in os.environ:
        if isinstance(p.annotation, typing._GenericAlias):
            type_name = p.annotation.__args__[len(p.annotation.__args__) - 1].__name__
            instance_type = p.annotation._name
            def _to_object(item):
                str_val = get_converter(type(item).__name__).convert_to(item)
                return get_converter(type_name).convert_from(str_val)
            if instance_type == "Dict":
                dict_values = get_converter("dict").convert_from(os.environ[env_var])
                value = {key: _to_object(item) for key, item in dict_values.items()}
            elif instance_type == "List":
                list_values = get_converter("list").convert_from(os.environ[env_var])
                value = [_to_object(item) for item in list_values]
        else:
            type_name = p.annotation.__name__
            value = get_converter(type_name).convert_from(os.environ[env_var])
    return value

def _get_service_provider_instance(p: Parameter):
    type_name = p.annotation.__name__
    return get_service_provider(type_name).get_instance()

def _write_output_to_environment(key, value):
    value_str = json.dumps(value) if isinstance(value, dict) or isinstance(value, list) else value
    with open(f"/tmp/{key}.txt", 'w') as out:
        out.write(value_str)
    logging.info(f"Done writing output:{key}")
    logging.info(value_str)
def task(inputs: set={}):
    def decorator_task(func):
        def execute_task(*args, skip_task_decoration=False):
            if skip_task_decoration or not _SETTINGS['enable_task_decoration']:
                return func(*args)

            try:
                sig = signature(func)
                if len(args) == 0:
                    args = []
                    for p in sig.parameters.values():
                        if p.name in inputs:
                            args.append(_get_argument(p))
                        else:
                            args.append(_get_service_provider_instance(p))
                result = func(*args)
                if result is not None and isinstance(result, dict):
                    for key, value in result.items():
                        _write_output_to_environment(key, value)
                return result
            except ValidationError as err:
                stop_on_failed(err.message, error_code=err.code, error_data=err.data, suggestion=err.suggestion)
            except grpc.RpcError as grpc_err:
                traceback.print_exc()
                error_message, error_code = unpack_grpc_error(grpc_err)
                stop_on_failed(error_message, error_code=error_code)
            except Exception as ex:
                traceback.print_exc()
                error_code = "INTERNAL_ERROR"
                error_message = f'{ex}'
                stop_on_failed(error_message, error_code=error_code)
        return execute_task
    return decorator_task
