from aiaengine.api import Client
from aiaengine.pipeline import register_converter, set_object_input

from .base import (
    IntegerConverter,
    FloatConverter,
    BooleanConverter,
    StringConverter,
    JSONConverter,
)

from .platform import (
    DATASET_OBJECT_TYPE,
    APP_OBJECT_TYPE,
    EVALUATION_OBJECT_TYPE,
    PREDICITON_OBJECT_TYPE,
    RECIPE_OBJECT_TYPE,
    FEATURE_SET_OBJECT_TYPE,
    MODEL_OBJECT_TYPE,
    TRAINING_OBJECT_TYPE,
)

from aiaengine.pipeline.converters.dataset import (
    PlatformDatasetLoader,
    register_dataset_converter,
)
from aiaengine.pipeline.converters.recipe import (
    PlatformRecipeLoader,
    register_recipe_converter,
)
from aiaengine.pipeline.converters.app import PlatformAppLoader, register_app_converter
from aiaengine.pipeline.converters.feature_set import (
    PlatformFeatureSetLoader,
    register_feature_set_converter,
)
from aiaengine.pipeline.converters.preprocessing_pipeline import (
    PlatformPreprocessingPipelineSpecLoader,
    register_preprocessing_pipeline_spec_converter,
)
from aiaengine.pipeline.converters.model import (
    PlatformEvaluationLoader,
    PlatformModelLoader,
    PlatformPredictionLoader,
    PlatformTrainingLoader,
    register_evaluation_converter,
    register_model_converter,
    register_prediction_converter,
    register_training_converter,
)
from aiaengine.pipeline.converters.artifact import (
    PlatformArtifactLoader,
    register_artifact_converter
)

def register_primitive_converters():
    register_converter("int", IntegerConverter())
    register_converter("float", FloatConverter())
    register_converter("bool", BooleanConverter())
    register_converter("str", StringConverter())
    register_converter("dict", JSONConverter())
    register_converter("list", JSONConverter())


def register_platform_object_converters(client: Client, spark, dask_cluster, dask_cuda_cluster):
    dataset_loader = PlatformDatasetLoader(client=client)
    register_dataset_converter(dataset_loader)

    if spark is not None:
        from aiaengine.pipeline.converters.spark import (
            SparkDatasetConverter,
            register_spark_dataset_converter,
        )

        register_spark_dataset_converter(
            SparkDatasetConverter(loader=dataset_loader, spark=spark)
        )

    if dask_cluster is not None:
        from aiaengine.pipeline.converters.dask import DaskDatasetConverter, register_dask_dataset_converter
        register_dask_dataset_converter(DaskDatasetConverter(dataset_loader))

    if dask_cuda_cluster is not None:
        from aiaengine.pipeline.converters.dask_cudf import DaskCudfDatasetConverter,  register_dask_cudf_dataset_converter
        register_dask_cudf_dataset_converter(DaskCudfDatasetConverter(dataset_loader))

    register_recipe_converter(
        PlatformRecipeLoader(client=client, dataset_loader=dataset_loader)
    )

    app_loader = PlatformAppLoader(client=client, dataset_loader=dataset_loader)
    register_app_converter(loader=app_loader)

    feature_set_loader = PlatformFeatureSetLoader(
        client=client, app_loader=app_loader, dataset_loader=dataset_loader
    )
    register_feature_set_converter(loader=feature_set_loader)

    preprocessing_pipeline_spec_loader = PlatformPreprocessingPipelineSpecLoader(
        client=client, feature_set_loader=feature_set_loader
    )
    register_preprocessing_pipeline_spec_converter(
        loader=preprocessing_pipeline_spec_loader
    )

    model_loader = PlatformModelLoader(
        client=client, feature_set_loader=feature_set_loader
    )
    register_model_converter(loader=model_loader)
    training_loader = PlatformTrainingLoader(client=client, model_loader=model_loader)
    register_training_converter(loader=training_loader)

    evaluation_loader = PlatformEvaluationLoader(
        client=client, training_loader=training_loader, dataset_loader=dataset_loader
    )
    register_evaluation_converter(loader=evaluation_loader)

    prediction_loader = PlatformPredictionLoader(
        client=client, training_loader=training_loader, dataset_loader=dataset_loader
    )
    register_prediction_converter(loader=prediction_loader)

    register_artifact_converter(loader=PlatformArtifactLoader(client=client))


def register_converters(api_client: Client, spark, dask_cluster, dask_cuda_cluster):
    register_primitive_converters()
    register_platform_object_converters(api_client, spark, dask_cluster, dask_cuda_cluster)


def set_dataset_input(arg_name, dataset_id, version):
    set_object_input(arg_name, DATASET_OBJECT_TYPE, dataset_id, version=version)


def set_recipe_input(arg_name, recipe_id, iteration):
    set_object_input(arg_name, RECIPE_OBJECT_TYPE, recipe_id, iteration=iteration)


def set_app_input(arg_name, app_id, dataset_version):
    set_object_input(arg_name, APP_OBJECT_TYPE, app_id,  dataset_version=dataset_version)


def set_feature_set_input(arg_name, feature_set_id, dataset_version):
    set_object_input(
        arg_name,
        FEATURE_SET_OBJECT_TYPE,
        feature_set_id,
        dataset_version=dataset_version,
    )


def set_model_input(arg_name, model_id, dataset_version):
    set_object_input(arg_name, MODEL_OBJECT_TYPE, model_id, dataset_version=dataset_version)


def set_training_input(arg_name, training_id, model_id):
    set_object_input(arg_name, TRAINING_OBJECT_TYPE, training_id, model_id=model_id)


def set_evaluation_input(arg_name, evaluation_id, model_id):
    set_object_input(arg_name, EVALUATION_OBJECT_TYPE, evaluation_id, model_id=model_id)


def set_prediction_input(arg_name, prediction_id, model_id):
    set_object_input(arg_name, PREDICITON_OBJECT_TYPE, prediction_id, model_id=model_id)
