import abc

from aiaengine import api
from aiaengine.api import app
from aiaengine.core import Dataset, App
from aiaengine.core.app import (
    AnomalyDetectionConfig,
    ClassificationConfig,
    ClusteringConfig,
    MLProblemType,
    RegressionConfig,
    TimeSeriesForecastingConfig
)
from aiaengine.pipeline import format_object_input, register_converter
from aiaengine.pipeline.converters.base import BasePlatformObjectConverter
from aiaengine.pipeline.converters.dataset import DatasetLoader
from aiaengine.pipeline.converters.platform import APP_OBJECT_TYPE
from google.protobuf import json_format


class AppLoader(abc.ABC):
    @abc.abstractmethod
    def load(self, id: str, dataset_version: int = -1) -> App:
        raise Exception("Not implemented")


class PlatformAppLoader(AppLoader):
    def __init__(self, client: api.Client, dataset_loader: DatasetLoader):
        self.client = client
        self.dataset_loader = dataset_loader

    def load(self, id: str, dataset_version: int = -1) -> App:
        app_info = self.client.apps.GetApp(app.GetAppRequest(id=id))
        # customer churn template does not have dataset when status is draft
        dataset = self.dataset_loader.load(
            id=app_info.dataset.id, version=dataset_version
        ) if app_info.dataset.id else None

        # new app has train dataset id and test dataset id
        train_data = self.dataset_loader.load(
            id=app_info.train_dataset_id, version=dataset_version
        ) if app_info.train_dataset_id else None
        test_data = (
            self.dataset_loader.load(
                id=app_info.test_dataset_id, version=dataset_version
            )
            if app_info.test_dataset_id
            else None
        )

        problem_type = app_info.problem_type
        problem_type_config_dict = {}
        if app_info.problem_type_config is not None:
            problem_type_config_dict = json_format.MessageToDict(
                app_info.problem_type_config
            )

        problem_type_config = None
        if problem_type == MLProblemType.CLASSIFICATION:
            problem_type_config = ClassificationConfig.from_dict(
                problem_type_config_dict
            )
        elif problem_type == MLProblemType.REGRESSION:
            problem_type_config = RegressionConfig.from_dict(
                problem_type_config_dict
            )
        elif problem_type == MLProblemType.CLUSTERING:
            problem_type_config = ClusteringConfig.from_dict(
                problem_type_config_dict
            )
        elif problem_type == MLProblemType.ANOMALY_DETECTION:
            problem_type_config = AnomalyDetectionConfig.from_dict(
                problem_type_config_dict
            )
        elif problem_type == MLProblemType.TIME_SERIES_FORECASTING:
            problem_type_config = TimeSeriesForecastingConfig.from_dict(
                problem_type_config_dict
            )

        return App(
            id=id,
            name=app_info.name,
            description=app_info.description,
            problem_type=app_info.problem_type,
            problem_type_config=problem_type_config,
            target_columns=app_info.target_columns,
            dataset=dataset,
            train_data=train_data,
            test_data=test_data,
            project_id=app_info.project.id,
            dataset_version=dataset_version,
            org_id=app_info.org.id,
            annotations=app_info.annotations,
            status=app_info.status,
        )


class AppConverter(BasePlatformObjectConverter):
    def __init__(self, loader: AppLoader):
        super().__init__(object_type=APP_OBJECT_TYPE)
        self.loader = loader

    def convert_from(self, value: str):
        json_data = self.parse_json(value)
        dataset_version = json_data.get("dataset_version", -1)
        return self.loader.load(json_data["id"], dataset_version)

    def convert_to(self, value: App):
        return format_object_input(value.id, dataset_version=value.dataset_version)


def register_app_converter(loader: AppLoader):
    register_converter(APP_OBJECT_TYPE, AppConverter(loader=loader))
