import abc
from aiaengine import api, io_utils
from aiaengine.api import dataset, file
from aiaengine.core import Schema, Dataset
from aiaengine.pipeline import format_object_input, register_converter
from aiaengine.pipeline.converters.base import BasePlatformObjectConverter
from aiaengine.pipeline.converters.platform import DATASET_OBJECT_TYPE

class DatasetLoader(abc.ABC):
    @abc.abstractmethod
    def load(self, id: str, version: int = 0) -> Dataset:
        pass

class PlatformDatasetLoader(DatasetLoader):
    def __init__(self, client: api.Client):
        self.client = client

    def load(self, id: str, version: int = 0) -> Dataset:
        dataset_response = self.client.datasets.GetDataset(dataset.GetDatasetRequest(id=id))
        # If dataset is importing, we don't have any dataset versions
        urls = []
        metadata = None
        if dataset_response.latest_version >= 0:
            # get list of dataset's files
            list_files_response = self.client.files.ListFiles(file.ListFilesRequest(dataset_id=id, version=version))
            urls = [f.url for f in list_files_response.files]
            # get metadata of dataset
            metadata_resp = self.client.datasets.GetMetadataUrl(
                dataset.GetMetadataUrlRequest(id=id, version=version)
            )
            metadata = io_utils.load_file_to_json(metadata_resp.url) if metadata_resp.url else None

        # get schema of dataset
        schema_resp = self.client.datasets.GetSchemaUrl(dataset.GetSchemaUrlRequest(id=id))
        print(f"Schema URL={schema_resp.url}")
        schema = Schema.from_file(schema_resp.url)
        # get sample data
        first5k_sample_url = self._get_sample_url(id, "first5K")
        first5k_dataset = Dataset(schema=schema, data_urls=[first5k_sample_url]) if first5k_sample_url else None
        random_sample_url = self._get_sample_url(id, "random")
        random_dataset = Dataset(schema=schema, data_urls=[random_sample_url]) if random_sample_url else None
        return Dataset(schema=schema, data_urls=urls,
                       id=id, version=version, metadata=metadata,
                       first_n_rows_sample_dataset=first5k_dataset,
                       random_n_rows_sample_dataset=random_dataset)

    def _get_sample_url(self, dataset_id, strategy):
        response = self.client.datasets.GetSampleDataUrl(
            dataset.GetSampleDataUrlRequest(id=dataset_id, strategy=strategy))
        return response.url if response.url != "" else None

class DatasetConverter(BasePlatformObjectConverter):
    def __init__(self, loader: DatasetLoader):
        super().__init__(object_type=DATASET_OBJECT_TYPE)
        self.loader = loader

    def convert_from(self, value: str):
        json_data = self.parse_json(value)
        return self.loader.load(id=json_data['id'], version=json_data['version'])

    def convert_to(self, value: Dataset):
        return format_object_input(value.id, version=value.version)

def register_dataset_converter(loader: DatasetLoader):
    register_converter(DATASET_OBJECT_TYPE, DatasetConverter(loader=loader))