import abc
import logging
import json
from aiaengine import api
from aiaengine.api import featureset
from aiaengine.core import FeatureSet
from aiaengine.core.preprocessing_pipeline import PreprocessingPipeline
from aiaengine.pipeline import format_object_input, register_converter
from aiaengine.pipeline.converters.app import AppLoader
from aiaengine.pipeline.converters.dataset import DatasetLoader
from aiaengine.pipeline.converters.base import BasePlatformObjectConverter
from aiaengine.pipeline.converters.platform import FEATURE_SET_OBJECT_TYPE


class FeatureSetLoader:
    @abc.abstractmethod
    def load(self, id: str, dataset_version: int = -1) -> FeatureSet:
        pass


class PlatformFeatureSetLoader(FeatureSetLoader):
    def __init__(
        self, client: api.Client, app_loader: AppLoader, dataset_loader: DatasetLoader
    ):
        self.client = client
        self.app_loader = app_loader
        self.dataset_loader = dataset_loader

    def load(self, id: str, dataset_version: int = -1) -> FeatureSet:
        feature_set_info = self.client.featuresets.GetFeatureSet(
            featureset.GetFeatureSetRequest(id=id)
        )
        app = self.app_loader.load(
            id=feature_set_info.app.id, dataset_version=dataset_version
        )

        preprocessing_pipeline = None
        preprocessed_train_dataset = None
        preprocessed_test_dataset = None
        if feature_set_info.preprocessing_pipeline_versions is not None:
            versions = [v for v in feature_set_info.preprocessing_pipeline_versions if v.version == dataset_version]
            if len(versions) > 0:
                pipeline_version = versions[0]
                logging.info(f"Loading pipeline from {pipeline_version.preprocessing_pipeline_url}")  
                preprocessing_pipeline = PreprocessingPipeline.load(
                    pipeline_version.preprocessing_pipeline_url
                )
                logging.info(f"Loading preprocessed_train_dataset of version {dataset_version}")
                preprocessed_train_dataset = self.dataset_loader.load(
                    pipeline_version.preprocessed_train_dataset_id, version=-1,
                )
                if pipeline_version.preprocessed_test_dataset_id:
                    logging.info(f"Loading preprocessed_test_dataset of version {dataset_version}")
                    preprocessed_test_dataset = self.dataset_loader.load(
                        pipeline_version.preprocessed_test_dataset_id, version=-1
                    )

        recommendation = []
        if feature_set_info.recommendations != '':
            recommendation = recommendation=json.loads(feature_set_info.recommendations)["recommendation"]
  
        return FeatureSet(
            id=id,
            name=feature_set_info.name,
            description=feature_set_info.description,
            feature_names=[feature_name for feature_name in feature_set_info.feature_names],
            preprocessing_pipeline=preprocessing_pipeline,
            preprocessed_train_dataset=preprocessed_train_dataset,
            preprocessed_test_dataset=preprocessed_test_dataset,
            app=app,
            dataset_version=dataset_version,
            recommendation=recommendation,
            preferred_feature_types=feature_set_info.preferred_feature_types
        )


class FeatureSetConverter(BasePlatformObjectConverter):
    """Convert JSON representation of feature set <-> FeatureSet object"""

    def __init__(self, loader: FeatureSetLoader):
        super().__init__(object_type=FEATURE_SET_OBJECT_TYPE)
        self.loader = loader

    def convert_from(self, value: str):
        json_data = self.parse_json(value)
        dataset_version = json_data.get("dataset_version", -1)
        return self.loader.load(json_data["id"], dataset_version=dataset_version)

    def convert_to(self, value: FeatureSet):
        return format_object_input(value.id, dataset_version=value.dataset_version)


def register_feature_set_converter(loader: FeatureSetLoader):
    register_converter(FEATURE_SET_OBJECT_TYPE, FeatureSetConverter(loader=loader))
