import abc
from google.protobuf import json_format

from aiaengine import api
from aiaengine.api import model
from aiaengine.core import (
    Model,
    Training,
    TrainingConfig,
    TrainedModel,
    EvaluationConfig,
    Evaluation,
)
from aiaengine.pipeline import register_converter

from aiaengine.pipeline.converters.dataset import DatasetLoader
from aiaengine.pipeline.converters.feature_set import FeatureSetLoader
from aiaengine.pipeline.converters.base import BasePlatformObjectConverter
from aiaengine.pipeline.converters.platform import (
    EVALUATION_OBJECT_TYPE,
    MODEL_OBJECT_TYPE,
    PREDICITON_OBJECT_TYPE,
    TRAINING_OBJECT_TYPE,
)
from aiaengine.core.model import Evaluation, Prediction, PredictionConfig


class ModelLoader:
    @abc.abstractmethod
    def load(self, id: str, dataset_version: int = -1) -> Model:
        raise Exception("Not implemented")


class TrainingLoader:
    @abc.abstractmethod
    def load(self, id: str, model_id: str) -> Training:
        raise Exception("Not implemented")


class EvaluationLoader:
    @abc.abstractmethod
    def load(self, id: str, model_id: str) -> Evaluation:
        raise Exception("Not implemented")


class PredictionLoader:
    @abc.abstractmethod
    def load(self, id: str, model_id: str) -> Prediction:
        raise Exception("Not implemented")


class PlatformModelLoader(ModelLoader):
    """Load model data from the platform"""

    def __init__(self, client: api.Client, feature_set_loader: FeatureSetLoader):
        self.client = client
        self.feature_set_loader = feature_set_loader

    def load(self, id: str, dataset_version: int = -1) -> Model:
        model_info = self.client.models.GetModel(model.GetModelRequest(id=id))
        feature_set = self.feature_set_loader.load(
            model_info.feature_set.id, dataset_version=dataset_version
        )

        return Model(
            id=id,
            name=model_info.name,
            template_id=model_info.template.id,
            feature_set=feature_set,
            dataset_version=dataset_version
        )


class PlatformTrainingLoader(TrainingLoader):
    """Load Training data from the platform"""

    def __init__(self, client: api.Client, model_loader: ModelLoader):
        self.client = client
        self.model_loader = model_loader

    def load(self, id: str, model_id: str) -> Model:
        """Load Training object from the platform
        Args:
            id (str): Training ID
            model_id (str): Model ID

        Returns:
            Training: [description]
        """
        training_info = self.client.models.GetTraining(
            model.GetTrainingRequest(id=model_id, training_id=id)
        )
        trained_model: TrainedModel = None
        if training_info.trained_model is not None:
            model_hyperparameters = json_format.MessageToDict(
                training_info.trained_model.hyperparameters
            )
            trained_model = TrainedModel(
                format=training_info.trained_model.format,
                url=training_info.trained_model.url,
                hyperparameters=model_hyperparameters,
            )

        hyperparameter_tuning_method = None
        hyperparameter_tuning_config = {}
        training_hyperparameters = {}
        if training_info.config is not None:
            if training_info.config.hyperparameter_tuning_method != "":
                hyperparameter_tuning_method = (
                    training_info.config.hyperparameter_tuning_method
                )
            hyperparameter_tuning_config = json_format.MessageToDict(
                training_info.config.hyperparameter_tuning_config
            )
            # Python protobuf convert all number to float, need convert some fields to integer
            hyperparameter_tuning_config = _parse_int_from_any(hyperparameter_tuning_config)

            training_hyperparameters = json_format.MessageToDict(
                training_info.config.hyperparameters
            )
            # Python protobuf convert all number to float, need convert some fields to integer
            training_hyperparameters = _parse_int_from_any(training_hyperparameters)

        return Training(
            id=id,
            model=self.model_loader.load(
                training_info.model.id, training_info.dataset_version
            ),
            config=TrainingConfig(
                hyperparameter_tuning_method=hyperparameter_tuning_method,
                hyperparameter_tuning_config=hyperparameter_tuning_config,
                hyperparameters=training_hyperparameters,
            ),
            trained_model=trained_model,
            dataset_version=training_info.dataset_version
        )

def _parse_int_from_any(value):
    """
        Loop through any value (it maybe dictionary, list) and try parse integer with rule int(v) == v.
        Example: value = {
            "max_depth": [2, 4, 8],
            "weights": ["uniform", "distance"],
            "learning_rate": [0.001, 0.01, 0.1, 0.5, 1.0]
        }
    """
    if isinstance(value, list):
        return [_try_parse_int(v) for v in value]
    elif isinstance(value, dict):
        return {k: _parse_int_from_any(v) for k, v in value.items()}
    else:
        return _try_parse_int(value)

def _try_parse_int(value):
    if isinstance(value, float):
        int_value = int(value)
        return int_value if int_value == value else value
    return value


class PlatformEvaluationLoader(EvaluationLoader):
    def __init__(
        self,
        client: api.Client,
        training_loader: TrainingLoader,
        dataset_loader: DatasetLoader,
    ):
        self.client = client
        self.training_loader = training_loader
        self.dataset_loader = dataset_loader

    def load(self, id: str, model_id: str) -> Evaluation:
        evaluation_info = self.client.models.GetEvaluation(
            model.GetEvaluationRequest(id=id, model_id=model_id)
        )
        training = self.training_loader.load(evaluation_info.training_id, model_id)
        #Old apps don't have test dataset id
        test_dataset = None
        if evaluation_info.test_dataset.id != "":
            test_dataset = self.dataset_loader.load(
                evaluation_info.test_dataset.id, evaluation_info.test_dataset.version
            )
        config = EvaluationConfig()
        evaluation = Evaluation(
            config=config, training=training, test_dataset=test_dataset, id=id
        )
        return evaluation


class PlatformPredictionLoader(PredictionLoader):
    def __init__(
        self,
        client: api.Client,
        training_loader: TrainingLoader,
        dataset_loader: DatasetLoader,
    ):
        self.client = client
        self.training_loader = training_loader
        self.dataset_loader = dataset_loader

    def load(self, id: str, model_id: str) -> Evaluation:
        prediction_info = self.client.models.GetPrediction(
            model.GetPredictionRequest(id=id, model_id=model_id)
        )
        training = self.training_loader.load(prediction_info.training.id, model_id)
        config = PredictionConfig(
            include_variables=prediction_info.config.include_variables
        )
        input_dataset = self.dataset_loader.load(
            prediction_info.input_dataset.id, prediction_info.input_dataset.version
        )
        return Prediction(
            config=config, training=training, input_dataset=input_dataset, id=id
        )


class ModelConverter(BasePlatformObjectConverter):
    """Convert from json <-> Model"""

    def __init__(self, loader: ModelLoader):
        super().__init__(object_type=MODEL_OBJECT_TYPE)
        self.loader = loader

    def convert_from(self, value: str):
        json_data = self.parse_json(value)
        dataset_version = json_data.get("dataset_version", -1)
        return self.loader.load(json_data["id"], dataset_version)


class TrainingConverter(BasePlatformObjectConverter):
    def __init__(self, loader: TrainingLoader):
        super().__init__(object_type=TRAINING_OBJECT_TYPE)
        self.loader = loader

    def convert_from(self, value: str):
        json_data = self.parse_json(value)
        return self.loader.load(json_data["id"], json_data["model_id"])


class EvaluationConverter(BasePlatformObjectConverter):
    def __init__(self, loader: EvaluationLoader):
        super().__init__(object_type=EVALUATION_OBJECT_TYPE)
        self.loader = loader

    def convert_from(self, value: str):
        json_data = self.parse_json(value)
        return self.loader.load(json_data["id"], json_data["model_id"])


class PredictionConverter(BasePlatformObjectConverter):
    def __init__(self, loader: PredictionLoader):
        super().__init__(object_type=PREDICITON_OBJECT_TYPE)
        self.loader = loader

    def convert_from(self, value: str):
        json_data = self.parse_json(value)
        return self.loader.load(json_data["id"], json_data["model_id"])


def register_model_converter(loader: ModelLoader):
    register_converter(MODEL_OBJECT_TYPE, ModelConverter(loader=loader))


def register_training_converter(loader: TrainingLoader):
    register_converter(TRAINING_OBJECT_TYPE, TrainingConverter(loader=loader))


def register_evaluation_converter(loader: EvaluationLoader):
    register_converter(EVALUATION_OBJECT_TYPE, EvaluationConverter(loader=loader))


def register_prediction_converter(loader: PredictionLoader):
    register_converter(PREDICITON_OBJECT_TYPE, PredictionConverter(loader=loader))
