import abc
from typing import Dict, List

from aiaengine import api, io_utils
from aiaengine.api import recipe
from aiaengine.core import Dataset, Recipe, PRIMARY_DATASET_ALIAS, is_primary_dataset
from aiaengine.pipeline import register_converter
from aiaengine.pipeline.converters import RECIPE_OBJECT_TYPE
from aiaengine.pipeline.converters.base import BasePlatformObjectConverter
from aiaengine.pipeline.converters.dataset import DatasetLoader

def get_committed_actions(recipe_info, iteration=-1):
    """Get committed actions of all iteration"""
    actions = []
    if iteration > 0:
        print("Get only actions of the current iteration (case commit actions)")
        it_info = recipe_info.iterations[iteration-1]
        actions_url = it_info.committed_actions_url
        if actions_url:
            actions = io_utils.load_file_to_json(actions_url)
            filter_valid_actions(actions, it_info.invalid_action_index)
        return actions

    for it_info in recipe_info.iterations:
        actions_url = it_info.committed_actions_url
        if actions_url:
            curr_actions = io_utils.load_file_to_json(actions_url)
            filter_valid_actions(curr_actions, it_info.invalid_action_index)
            actions.extend(curr_actions)
    return actions

def filter_valid_actions(full_actions_list: List[dict], invalid_action_index: int):
    if invalid_action_index >= 0:
        for full_actions in full_actions_list:
            actions = full_actions['solution']['actions']
            full_actions['solution']['actions'] = actions[:invalid_action_index]


class RecipeLoader:
    @abc.abstractmethod
    def load(self, id: str, iteration: int) -> Recipe:
        pass

class PlatformRecipeLoader(RecipeLoader):
    def __init__(self, client: api.Client, dataset_loader: DatasetLoader):
        self.client = client
        self._dataset_loader = dataset_loader

    def load(self, id: str, iteration: int=-1) -> Recipe:
        recipe_info = self.client.recipes.GetRecipe(recipe.GetRecipeRequest(id=id))
        actions = get_committed_actions(recipe_info, iteration)
        last_iteration = max(it.iteration for it in recipe_info.iterations)
        if iteration == -1:
            iteration = last_iteration
        input_datasets = self._get_input_datasets(recipe_info, iteration)
        return Recipe(actions=actions, input_datasets=input_datasets,
                     last_iteration=last_iteration, id=id, name=recipe_info.name, project_id=recipe_info.project.id)

    def _get_input_datasets(self, recipe_info, iteration: int) -> Dict[str, Dataset]:
        """Returns dictionary of input datasets"""
        input_ds = {ds.alias if not is_primary_dataset(ds.alias) else PRIMARY_DATASET_ALIAS: (ds.id, ds.version) \
                for ds in recipe_info.datasets}
        # if iteration > 1, the primary dataset will be at last iteration
        if iteration > 1:
            ds_id = recipe_info.iterations[iteration-1].dataset_id
            input_ds[PRIMARY_DATASET_ALIAS] = (ds_id, 0)
        return {alias: self._dataset_loader.load(id=value[0], version=value[1]) for alias, value in input_ds.items()}

class RecipeConverter(BasePlatformObjectConverter):
    def __init__(self, loader: RecipeLoader):
        super().__init__(object_type=RECIPE_OBJECT_TYPE)
        self.loader = loader

    def convert_from(self, value: str):
        json_data = self.parse_json(value)
        return self.loader.load(json_data['id'], json_data.get('iteration', -1))

def register_recipe_converter(loader: RecipeLoader):
    register_converter(RECIPE_OBJECT_TYPE, RecipeConverter(loader=loader))