""" Resolve Spark DataFrame """
from functools import reduce
from threading import Lock
from typing import Dict, Union

from aiaengine.core import Dataset, Schema, DEFAULT_SAMPLE_SIZE
from aiaengine.pipeline import Converter, register_converter
from aiaengine.pipeline.converters.dataset import DatasetConverter, DatasetLoader
from pyspark.sql import DataFrame
from pyspark.sql.session import SparkSession

from .dtype_mappings import spark_type_agrees_with_data_type


class SparkDataset(Dataset):
    def __init__(self, dataset: Dataset, spark: SparkSession, df: DataFrame = None):
        super().__init__(schema=dataset.schema,
                         data_urls=dataset.data_urls,
                         id=dataset.id,
                         version=dataset.version,
                         metadata=dataset.metadata,
                         first_n_rows_sample_dataset=dataset.first_n_rows_sample,
                         random_n_rows_sample_dataset=dataset.random_n_rows_sample
                         )
        self._first_n_rows_sample_spark_ds = None if not isinstance(dataset.first_n_rows_sample,
                                                                    SparkDataset) else dataset.first_n_rows_sample
        self._random_n_rows_sample_spark_ds = None if not isinstance(dataset.random_n_rows_sample,
                                                                     SparkDataset) else dataset.random_n_rows_sample
        self.spark = spark
        self._df = df
        self.sample_size = DEFAULT_SAMPLE_SIZE
        self._get_df_lock = Lock()
        self._first_n_sample_lock = Lock()
        self._random_sample_lock = Lock()

    @staticmethod
    def from_dataframe(df: DataFrame, schema: Union[Dict, Schema], spark_session: SparkSession):
        if not isinstance(schema, Schema):
            schema = Schema.from_dict(schema)
        return SparkDataset(Dataset(schema=schema, data_urls=None), spark_session, df=df)

    def get_df(self, keep_order=False):
        if self._df is None:
            with self._get_df_lock:
                if not keep_order:
                    self._df = self.spark.read.parquet(*self.data_urls)
                else:
                    # for timeseries data, must read file in sequences
                    # because spark does not keep ordering of files when reading
                    list_dfs = [self.spark.read.parquet(file_path) for file_path in self.data_urls]
                    self._df = reduce(lambda df1, df2: df1.union(df2), list_dfs)

                # validate column's type in schema vs dataframe
                _check_dataframe_schema_alignment(self._df, self.schema)
        return self._df

    @property
    def first_n_rows_sample(self):
        """ First N row sample """
        if self._first_n_rows_sample_spark_ds is None:
            with self._first_n_sample_lock:
                self._first_n_rows_sample_spark_ds = SparkDataset.from_dataframe(
                    df=self._get_first_n_row_df(),
                    schema=self.schema,
                    spark_session=self.spark) if self._get_first_n_row_df() else None
        return self._first_n_rows_sample_spark_ds

    @property
    def random_n_rows_sample(self):
        """ Random N row sample """
        if self._random_n_rows_sample_spark_ds is None:
            with self._random_sample_lock:
                self._random_n_rows_sample_spark_ds = SparkDataset.from_dataframe(
                    df=self._get_random_sample_df(),
                    schema=self.schema,
                    spark_session=self.spark) if self._get_random_sample_df() else None
        return self._random_n_rows_sample_spark_ds

    def _get_first_n_row_df(self):
        """
        Returns the first N rows Spark's DataFrame of dataset
        """
        if self._first_n_rows_sample_dataset is not None:
            first_n_df = SparkDataset(self._first_n_rows_sample_dataset, self.spark).get_df()
            return first_n_df.limit(self.sample_size)
        else:
            # In some case, sample data is not ready yet, return None
            return None

    def _get_random_sample_df(self):
        """
        Return a random sampling Spark's DataFrame of dataset
        """
        if self._random_n_rows_sample_dataset is not None:
            # load full random n row sample from underlying data, not limit it, client code will handle the size
            return SparkDataset(self._random_n_rows_sample_dataset, self.spark).get_df()
        else:
            # In some case, sample data is not ready yet, return None
            return None


def _wrap_col_name(col_name: str):
    return '`' + col_name.replace('`', '``') + '`'


def _check_dataframe_schema_alignment(main_df: DataFrame, schema: Schema):
    """
    Check if a Spark dataframe is aligned with input schema
    :param main_df: Spark dataframe
    :param schema_json: input schema
    :raises: AssertionError if there is mismatch between dataframe and schema
    """
    df_col_names = set(main_df.columns)
    schema_columns = schema.columns

    schema_col_names = [x.name for x in schema_columns]
    schema_set = set(schema_col_names)
    diff_set = schema_set.difference(df_col_names)
    if len(diff_set) > 0:
        raise AssertionError(f'Columns({diff_set}) differ between schema and dataset! '
                             f'Check schema and/or dataset!')

    # Verify column semantic type matches Spark dtype
    for col_info in schema_columns:
        column_name = col_info.name
        data_type = col_info.data_type
        dtype_in_df = main_df.select(_wrap_col_name(column_name)).schema[0].dataType
        assert spark_type_agrees_with_data_type(dtype_in_df, data_type), (
            f'Column "{column_name}" has dtype '
            f'"{dtype_in_df}" in Spark dataframe, but schema says data type is '
            f'"{data_type}"'
        )


class SparkDatasetConverter(DatasetConverter):
    def __init__(self, loader: DatasetLoader, spark: SparkSession):
        super().__init__(loader=loader)
        self.spark = spark

    def convert_from(self, val: str) -> SparkDataset:
        dataset = super().convert_from(val)
        return SparkDataset(dataset=dataset, spark=self.spark)

    def convert_to(self, val: SparkDataset) -> str:
        return super().convert_to(val)


def register_spark_dataset_converter(converter: Converter):
    register_converter('SparkDataset', converter)
