""" Platform Dataset Writer """
import logging
from typing import Dict, List
from abc import abstractclassmethod
from aiaengine.api import Client, file, dataset
from aiaengine.core import Dataset, Schema
from aiaengine.io_utils import upload_s3
from aiaengine.util import generate_file_upload_url

FIRST_5K_STRATEGY = 'first5K'
RANDOM_STRATEGY = 'random'
FIRST_1K_JSON_STRATEGY = 'first1KJson'

DATASET_TYPE_DEFAULT = ''
DATASET_TYPE_INTERMEDIATE = 'intermediate'


class DatasetWriter:
    @abstractclassmethod
    def write(self,
              schema: Schema,
              dataframe: any,
              project_id: str,
              name: str,
              dataset_type: str = DATASET_TYPE_DEFAULT,
              description: str = '',
              metadata: Dict = None,
              sample_data: Dict[str, any] = None) -> any:
        """ Write dataset to platform or any environment

        Args:
            schema: schema of dataset
            dataframe: any Dataframe format, such as Spark, Pandas or Dask
            metadata: metadata of dataset
            sample_data: Dictionary of any DataFrame format of sample data, key is sample strategy

        Returns:
            any: [description]
        """


class PlatformDatasetWriter(DatasetWriter):
    """  """

    def __init__(self, client: Client) -> None:
        self._client = client

    @abstractclassmethod
    def save_dataframe(self, project_id, dataframe) -> List[str]:
        """ Save DataFrame in any format to platform"""

    def write(self,
              schema: Schema,
              dataframe: any,
              project_id: str,
              name: str,
              dataset_type: str,
              description: str = '',
              metadata: Dict = None,
              sample_data: Dict[str, any] = None) -> any:
        # upload schema to platform
        logging.debug('Start saving schema')
        schema_url = generate_file_upload_url(self._client,
                                            project_id,
                                            name="schema.json",
                                            content_type="application/json")

        upload_s3(schema.to_dict(), schema_url)
        # save dataframe
        logging.debug(f'Start writing dataframe...')
        data_urls = self.save_dataframe(project_id, dataframe)
        # create dataset in platform
        request = dataset.CreateDatasetRequest(name=name,
                                            description=description,
                                            project_id=project_id,
                                            type=dataset_type,
                                            urls=data_urls,
                                            file_format='{"file_extension": "parquet"}',
                                            schema_url=schema_url)
        new_dataset = self._client.datasets.CreateDataset(request)
        dataset_id = new_dataset.id
        dataset_version = new_dataset.latest_version
        logging.info(
            f'New dataset is created, id={dataset_id}, version={dataset_version}')
        # upload metadata
        if metadata is not None:
            logging.debug('Start saving metadata')
            metadata_url = generate_file_upload_url(self._client,
                                            project_id,
                                            name="metadata.json",
                                            content_type="application/json")
            upload_s3(metadata, metadata_url)
            request = dataset.UpdateAnalysisResultRequest(
                id=dataset_id, version=0,
                result=dataset.AnalysisResult(
                    row_count=metadata["row_count"],
                    column_count=len(metadata["columns"]),
                    file_size=metadata["analysis_stats"]["file_size"],
                    analysis_time=metadata["analysis_stats"]["analysis_completion_time"]
                ),
                metadata_url=metadata_url,
            )
            self._client.datasets.UpdateAnalysisResult(request)

        # upload sample data
        first_n_row_sample_dataset = None
        random_n_row_sample_dataset = None
        if sample_data is not None:
            for strategy, sample in sample_data.items():
                upload_url = self._update_sample_data(
                    project_id, dataset_id, sample, strategy)
                if FIRST_5K_STRATEGY == strategy:
                    first_n_row_sample_dataset = Dataset(schema,
                                                        data_urls=[upload_url])
                if RANDOM_STRATEGY == strategy:
                    random_n_row_sample_dataset = Dataset(schema,
                                                        data_urls=[upload_url])

        return Dataset(schema, data_urls,
                       id=dataset_id,
                       version=dataset_version,
                       metadata=metadata,
                       first_n_rows_sample_dataset=first_n_row_sample_dataset,
                       random_n_rows_sample_dataset=random_n_row_sample_dataset)

    def _update_sample_data(self, project_id, dataset_id, sample, strategy):
        upload_url = ''
        if strategy in [FIRST_5K_STRATEGY, RANDOM_STRATEGY]:
            urls = self.save_dataframe(project_id, sample)
            upload_url = urls[0]
        else:
            upload_url = generate_file_upload_url(
                self._client, project_id, 'sample.json', 'application/json')
            upload_s3(sample, upload_url)
        logging.info(
            f'Uploaded sample data: strategy={strategy}, upload={upload_url}')
        self._client.datasets.UpdateSampleDataUrl(dataset.UpdateSampleDataUrlRequest(
            id=dataset_id,
            strategy=strategy,
            url=upload_url
        ))
        return upload_url
