import dataclasses
from aiaengine.api import Client
from aiaengine.api.feature_set_pb2 import UpdateFeatureSetRequest, CreateFeatureSetRequest, PreprocessingPipelineVersion, GetFeatureSetRequest
from aiaengine.core import FeatureSet
from aiaengine.core.preprocessing_pipeline import PreprocessingPipeline
from aiaengine.pipeline.providers import ServiceProvider
from aiaengine.waiter import wait_featureset_processed
import json
import logging

class FeatureSetWriter:
    def write(self, feature_set: FeatureSet, timeout: int = 0) -> FeatureSet:
        return feature_set


class PlatformFeatureSetWriter(FeatureSetWriter):
    def __init__(self, client: Client):
        self.client = client

    def write(self, feature_set: FeatureSet, timeout: int = 0) -> FeatureSet:
        if feature_set.id is None:
            response = self.client.featuresets.CreateFeatureSet(
                CreateFeatureSetRequest(
                    name=feature_set.name,
                    description=feature_set.description,
                    app_id=feature_set.app.id,
                    selected_features=feature_set.feature_names,
                    preferred_feature_types=feature_set.preferred_feature_types
                )
            )
            feature_set_id = response.id
            if timeout > 0:
                # wait for feature set processed
                updated_feature_set = wait_featureset_processed(self.client, feature_set_id, timeout=timeout)

                preprocessing_pipeline = None
                preprocessed_train_dataset = None
                preprocessed_test_dataset = None
                dataset_version = feature_set.app.dataset_version
                if updated_feature_set.preprocessing_pipeline_versions is not None:
                    versions = [v for v in updated_feature_set.preprocessing_pipeline_versions if v.version == dataset_version]
                    if len(versions) > 0:
                        pipeline_version = versions[0]
                        logging.info(f"Loading pipeline from {pipeline_version.preprocessing_pipeline_url}")  
                        preprocessing_pipeline = PreprocessingPipeline.load(
                            pipeline_version.preprocessing_pipeline_url
                        )
                        logging.info(f"Loading preprocessed_train_dataset of version {dataset_version}")
                        preprocessed_train_dataset = self.dataset_loader.load(
                            pipeline_version.preprocessed_train_dataset_id, version=-1,
                        )
                        if pipeline_version.preprocessed_test_dataset_id:
                            logging.info(f"Loading preprocessed_test_dataset of version {dataset_version}")
                            preprocessed_test_dataset = self.dataset_loader.load(
                                pipeline_version.preprocessed_test_dataset_id, version=-1
                            )
                recommendations = updated_feature_set.recommendations
                if recommendations == '':
                    recommendations = '{"recommendation":[]}'
                recommendation=json.loads(recommendations)["recommendation"]

                return dataclasses.replace(feature_set,
                    id=feature_set_id,
                    preprocessing_pipeline=preprocessing_pipeline,
                    preprocessed_train_dataset=preprocessed_train_dataset,
                    preprocessed_test_dataset=preprocessed_test_dataset,
                    recommendation=recommendation)

            return dataclasses.replace(feature_set, id=feature_set_id)
        else:
            preprocessing_pipeline_version = PreprocessingPipelineVersion(
                version=feature_set.dataset_version,
                preprocessing_pipeline_url=feature_set.preprocessing_pipeline.url,
                preprocessed_train_dataset_id=feature_set.preprocessed_train_dataset.id,
                preprocessing_pipeline_spec_id=feature_set.preprocessing_pipeline.spec.id,
            )
            if feature_set.preprocessed_test_dataset is not None:
                preprocessing_pipeline_version.preprocessed_test_dataset_id = feature_set.preprocessed_test_dataset.id

            response = self.client.featuresets.UpdateFeatureSet(
                UpdateFeatureSetRequest(
                    id=feature_set.id,
                    name=feature_set.name,
                    description=feature_set.description,
                    selected_features=feature_set.feature_names,
                    preprocessing_pipeline_version=preprocessing_pipeline_version,
                )
            )
        return feature_set


class FeatureSetWriterProvider(ServiceProvider):
    def __init__(self, client: Client):
        self.writer = PlatformFeatureSetWriter(client=client)

    def get_instance(self) -> FeatureSetWriter:
        return self.writer
