
from aiaengine.api import Client
from aiaengine.api import dataset as api_dataset
from aiaengine.api import model as api_model
from aiaengine.core.model import PredictionResult
from aiaengine.waiter import wait_model_trained
from aiaengine.util import upload_model
from aiaengine.core import Model, TrainingConfig, TrainingResult, EvaluationResult
from aiaengine.pipeline import ServiceProvider
from google.protobuf import struct_pb2, json_format

import dataclasses

class ModelWriter:
    def write(self, model: Model, training_config: TrainingConfig, timeout: int = 0) -> Model:
        return model

class TrainingResultWriter:
    def write(self, training_result: TrainingResult) -> TrainingResult:
        return training_result


class EvaluationResultWriter:
    def write(self, evaluation_result: EvaluationResult) -> EvaluationResult:
        return evaluation_result


class PredictionResultWriter:
    def write(self, prediction_result: PredictionResult) -> PredictionResult:
        return prediction_result

class PlatformModelWriter(ModelWriter):
    def __init__(self, client: Client):
        self.client = client

    def write(self, model: Model, training_config: TrainingConfig, timeout: int = 0) -> Model:
        api_training_config = api_model.TrainingConfig(
            hyperparameters=json_format.ParseDict(training_config.hyperparameters, struct_pb2.Struct()) if training_config.hyperparameters else None,
            hyperparameter_tuning_method=training_config.hyperparameter_tuning_method,
            hyperparameter_tuning_config=json_format.ParseDict(training_config.hyperparameter_tuning_config, struct_pb2.Struct()) if training_config.hyperparameter_tuning_config else None,
        ) if training_config else None

        # We get train dataset version instead of input dataset version
        train_dataset = self.client.datasets.GetDataset(
            api_dataset.GetDatasetRequest(id=model.feature_set.app.train_data.id)
        )
        if model.feature_set.app.dataset_version == -1:
            dataset_version = train_dataset.latest_version
        else:
            if model.feature_set.app.dataset_version > train_dataset.latest_version:
                dataset_version = train_dataset.latest_version
            else:
                dataset_version = model.feature_set.app.dataset_version
        response = self.client.models.CreateModel(
            api_model.CreateModelRequest(
                app_id=model.feature_set.app.id,
                feature_set_id=model.feature_set.id,
                dataset_version=dataset_version,
                name=model.name,
                template_id=model.template_id,
                training_config=api_training_config,
                continuous_learning=model.continuous_learning,
            )
        )
        model_id = response.id
        if timeout > 0:
            wait_model_trained(self.client, model_id, dataset_version, timeout)

        return dataclasses.replace(model, id=model_id)

    
class PlatformTrainingResultWriter(TrainingResultWriter):
    def __init__(self, client: Client):
        self.client = client

    def write(self, training_result: TrainingResult) -> TrainingResult:
        # upload trained model to S3
        upload_model(
            path=training_result.trained_model.url,
            url=training_result.training.trained_model.url,
        )
        training_stats = struct_pb2.Struct()
        json_format.ParseDict(training_result.training_stats, training_stats)
        hyperparameters = struct_pb2.Struct()
        json_format.ParseDict(
            training_result.trained_model.hyperparameters, hyperparameters
        )
        self.client.models.UpdateTrainingResult(
            api_model.UpdateTrainingResultRequest(
                id=training_result.training.model.id,
                training_id=training_result.training.id,
                result="{}",
                model_format=training_result.trained_model.format,
                hyperparameters=hyperparameters,
                training_stats=training_stats,
            )
        )
        return training_result


class PlatformEvaluationResultWriter(EvaluationResultWriter):
    def __init__(self, client: Client):
        self.client = client

    def write(self, evaluation_result: EvaluationResult) -> EvaluationResult:
        summary = struct_pb2.Struct()
        json_format.ParseDict(evaluation_result.summary, summary)
        details = struct_pb2.Struct()
        json_format.ParseDict(evaluation_result.details, details)
        result = api_model.EvaluationResult(
            summary=summary,
            details=details
        )
        evaluation_stats = struct_pb2.Struct()
        json_format.ParseDict(evaluation_result.evaluation.evaluation_stats, evaluation_stats)
        self.client.models.UpdateEvaluation(
            api_model.UpdateEvaluationRequest(
                id=evaluation_result.evaluation.id,
                model_id=evaluation_result.evaluation.training.model.id,
                result=result,
                evaluation_stats=evaluation_stats,
            )
        )
        return evaluation_result


class PlatformPredictionResultWriter(PredictionResultWriter):
    def __init__(self, client: Client):
        self.client = client

    def write(self, prediction_result: PredictionResult) -> PredictionResult:
        self.client.models.UpdatePrediction(
            api_model.UpdatePredictionRequest(
                id=prediction_result.prediction.id,
                model_id=prediction_result.prediction.training.model.id,
                output_dataset=api_dataset.RefDataset(
                    id=prediction_result.output_dataset.id,
                    version=prediction_result.output_dataset.version,
                ),
            )
        )
        return prediction_result


class ModelWriterProvider(ServiceProvider):
    def __init__(self, writer: ModelWriter):
        self.writer = writer

    def get_instance(self) -> ModelWriter:
        return self.writer

class TrainingResultWriterProvider(ServiceProvider):
    def __init__(self, writer: TrainingResultWriter):
        self.writer = writer

    def get_instance(self) -> TrainingResultWriter:
        return self.writer


class EvaluationResultWriterProvider(ServiceProvider):
    def __init__(self, writer: EvaluationResultWriter):
        self.writer = writer

    def get_instance(self) -> EvaluationResultWriter:
        return self.writer


class PredictionResultWriterProvider(ServiceProvider):
    def __init__(self, writer: PredictionResultWriter):
        self.writer = writer

    def get_instance(self) -> PredictionResultWriter:
        return self.writer
