import dataclasses
from aiaengine.api import Client, preprocessing_pipeline as pp
from aiaengine import io_utils
from aiaengine.core import PreprocessingPipeline, PreprocessingPipelineSpec
from aiaengine.pipeline.providers import ServiceProvider
from aiaengine.util import generate_file_upload_url
from google.protobuf import struct_pb2
from google.protobuf import json_format
import json
import uuid
import os


class PreprocessingPipelineSpecWriter:
    def write(
        self, pipeline_spec: PreprocessingPipelineSpec
    ) -> PreprocessingPipelineSpec:
        return pipeline_spec


class PreprocessingPipelineWriter:
    def generate_output_url(self, project_id: str) -> str:
        pipeline_output_url = f".output/pipeline-{str(uuid.uuid4())}"
        if not os.path.exists(pipeline_output_url):
            os.makedirs(pipeline_output_url)
        return pipeline_output_url

    def write(self, pipeline: PreprocessingPipeline) -> PreprocessingPipeline:
        return pipeline


class PlatformPreprocessingPipelineSpecWriter(PreprocessingPipelineSpecWriter):
    def __init__(self, client: Client):
        self._client = client

    def write(
        self, pipeline_spec: PreprocessingPipelineSpec
    ) -> PreprocessingPipelineSpec:
        protobuf_spec = struct_pb2.Struct()
        json_format.Parse(json.dumps(pipeline_spec.spec), protobuf_spec)

        response = self._client.preprocessing_pipelines.CreatePreprocessingPipelineSpec(
            pp.CreatePreprocesingPipelineSpecRequest(
                name=pipeline_spec.name,
                description=pipeline_spec.description,
                spec=protobuf_spec,
                feature_set_id=pipeline_spec.feature_set.id,
            )
        )

        return dataclasses.replace(pipeline_spec, id=response.id)


class PlatformPreprocessingPipelineWriter(PreprocessingPipelineWriter):
    def __init__(self, client: Client):
        self._client = client

    def generate_output_url(self, project_id: str) -> str:
        return generate_file_upload_url(
            client=self._client, project_id=project_id, name="pipeline"
        )

    def write(self, pipeline: PreprocessingPipeline) -> PreprocessingPipeline:
        # write metadata
        pipeline.save()
        if not pipeline.url.startswith("s3://"):
            url = self.generate_output_url(
                project_id=pipeline.spec.feature_set.app.project_id
            )
            io_utils.upload_folder_to_s3(pipeline.url, url)
            return dataclasses.replace(pipeline, url=url)

        return pipeline


class PreprocessingPipelineSpecWriterProvider(ServiceProvider):
    def __init__(self, writer: PreprocessingPipelineSpecWriter) -> None:
        self.writer = writer

    def get_instance(self) -> any:
        return self.writer


class PreprocessingPipelineWriterProvider(ServiceProvider):
    def __init__(self, writer: PreprocessingPipelineWriter) -> None:
        self.writer = writer

    def get_instance(self) -> any:
        return self.writer
