import json
from aiaengine.api import Client, recipe_pb2 as recipe_api
from aiaengine.core import Recipe
from aiaengine.pipeline.providers import ServiceProvider
from aiaengine.waiter import wait_until

class RecipeWriter:
    def write(self, recipe: Recipe, complete: bool = False, commit: bool = True, output_dataset_name: str = None, timeout: int = 600) -> Recipe:
        return recipe

class PlatformRecipeWriter(RecipeWriter):
    def __init__(self, client: Client):
        self.client = client

    def write(self, recipe: Recipe, complete: bool = False, commit: bool = True, output_dataset_name: str = None, timeout: int = 600) -> Recipe:
        if recipe.id is None:
            recipe_data = self.client.recipes.CreateRecipe(
                recipe_api.CreateRecipeRequest(
                    name=recipe.name,
                    datasets=[
                        recipe_api.InputDataset(
                            id=ds.id,
                            version=int(ds.version),
                            alias=alias
                        ) for alias, ds in recipe.input_datasets.items()
                    ],
                    disable_recommended_actions=True,
                )
            )
            recipe.id = recipe_data.id
            add_response = self.client.recipes.AddActions(
                recipe_api.AddActionsRequest(
                    id=recipe.id,
                    iteration=recipe.last_iteration,
                    actions=json.dumps(recipe.actions)
                )
            )
            if add_response.error:
                raise RuntimeError(add_response.error)

            if commit:
                self.client.recipes.CommitActions(
                recipe_api.CommitActionsRequest(
                        id=recipe.id,
                        iteration=recipe.last_iteration,
                    )
                )
                current_iteration = recipe.last_iteration
                # wait until actions processed
                def recipe_processed():
                    data = self.client.recipes.GetRecipe(recipe_api.GetRecipeRequest(id=recipe.id))
                    iteration = data.iterations[current_iteration-1]
                    if iteration.step == "file_format":
                        if iteration.status == "success":
                            return True
                        if iteration.status == "failed":
                            raise RuntimeError(f"Run recipe {recipe.id} failed")

                wait_until(recipe_processed, timeout=timeout)

            if complete:
                # complete the recipe
                response = self.client.recipes.CompleteRecipe(
                    recipe_api.CompleteRecipeRequest(
                        id=recipe.id,
                        dataset_name=output_dataset_name
                    )
                )
                recipe.output_dataset_id = response.dataset_id
            else:
                if commit:
                    response = self.client.recipes.GetRecipe(recipe_api.GetRecipeRequest(id=recipe.id))
                    iteration = response.iterations[len(response.iterations)-1]
                    recipe.output_dataset_id = iteration.dataset_id

            return recipe


class RecipeWriterProvider(ServiceProvider):
    def __init__(self, client: Client):
        self.client = client

    def get_instance(self) -> PlatformRecipeWriter:
        return PlatformRecipeWriter(client=self.client)