from typing import Dict
from pyspark.sql import DataFrame
from pyspark.sql.session import SparkSession

from aiaengine import io_utils
from aiaengine.util import generate_file_upload_url
from aiaengine.api import Client
from aiaengine.core import Schema, Dataset
from aiaengine.pipeline import ServiceProvider
from aiaengine.pipeline.converters.spark import SparkDataset

from aiaengine.pipeline.providers.dataset_reader import DatasetReader
from aiaengine.pipeline.providers.dataset_writer import DatasetWriter, PlatformDatasetWriter

from aiaengine.pipeline.providers.dataset_writer import DATASET_TYPE_DEFAULT


class SparkDatasetReader(DatasetReader):
    def __init__(self, spark: SparkSession) -> None:
        self.spark = spark

    def to_spark_dataset(self, dataset: Dataset) -> SparkDataset:
        return SparkDataset(dataset=dataset, spark=self.spark)

    def read(self, dataset: Dataset) -> DataFrame:
        return self.to_spark_dataset(dataset).get_df()


class SparkDatasetWriter(DatasetWriter):
    pass


class PlatformSparkDatasetWriter(PlatformDatasetWriter, SparkDatasetWriter):
    """A SparkDatasetWriter implementation that uses the Platform SDK to write data
       and create a new dataset
    """

    def __init__(self, client: Client, spark: SparkSession):
        PlatformDatasetWriter.__init__(self, client)
        self.spark = spark

    def save_dataframe(self, project_id, dataframe: DataFrame) -> SparkDataset:
        # generate url
        parquet_url = generate_file_upload_url(self._client, project_id)
        # write dataframe
        dataframe.write.mode("overwrite").parquet(parquet_url)
        # list out all parts of parquet
        data_urls = io_utils.list_urls(parquet_url, "parquet")
        return data_urls

    def write(self,
              schema: Schema,
              dataframe: any,
              project_id: str,
              name: str,
              dataset_type: str = DATASET_TYPE_DEFAULT,
              description: str = '',
              metadata: Dict = None,
              sample_data: Dict[str, any] = None) -> any:
        new_dataset = PlatformDatasetWriter.write(self,
                                                  schema=schema,
                                                  dataframe=dataframe,
                                                  project_id=project_id,
                                                  name=name,
                                                  dataset_type=dataset_type,
                                                  description=description,
                                                  metadata=metadata,
                                                  sample_data=sample_data)
        return SparkDataset(dataset=new_dataset, spark=self.spark)


class SparkDatasetWriterProvider(ServiceProvider):
    def __init__(self, writer: SparkDatasetWriter):
        self.writer = writer

    def get_instance(self) -> SparkDatasetWriter:
        return self.writer


class SparkDatasetReaderProvider(ServiceProvider):
    def __init__(self, reader: SparkDatasetReader):
        self.reader = reader

    def get_instance(self) -> SparkDatasetReader:
        return self.reader

class SparkSessionProvider(ServiceProvider):
    def __init__(self, spark: SparkSession):
        self.spark = spark

    def get_instance(self) -> SparkSession:
        return self.spark
