import os
import logging
from aiaengine.api.client import Client
from aiaengine.pipeline.converters import register_converters
from aiaengine.pipeline.providers import register_service_provider, register_service_providers
from aiaengine.pipeline import enable_task_decoration
from aiaengine.pipeline.providers.dataset_reader import DATAFRAME_FORMAT_DASK, DATAFRAME_FORMAT_SPARK, DATAFRAME_FORMAT_PANDAS, DATAFRAME_FORMAT_CUDF, DATAFRAME_FORMAT_DASK_CUDF
from aiaengine.pipeline.providers.cluster_factory import CLUSTER_TYPE_DASK, CLUSTER_TYPE_SPARK, CLUSTER_TYPE_NONE, CLUSTER_TYPE_DASK_CUDA
from aiaengine.pipeline.providers.pandas import PandasDatasetReader, PandasDatasetReaderProvider

def set_hadoop_spark_parameters(spark_session):
    """
    Set paramrters for hadoop
    :param spark_session:
    :return:  None
    """
    hadoop_conf = spark_session.sparkContext._jsc.hadoopConfiguration()
    hadoop_conf.set('fs.s3a.endpoint', os.getenv('AWS_ENDPOINT', ''))
    hadoop_conf.set('fs.s3a.access.key', os.getenv('AWS_ACCESS_KEY_ID', ''))
    hadoop_conf.set('fs.s3a.secret.key', os.getenv(
        'AWS_SECRET_ACCESS_KEY', ''))
    hadoop_conf.set('fs.s3a.path.style.access', 'true')
    hadoop_conf.set('fs.s3.impl', 'org.apache.hadoop.fs.s3a.S3AFileSystem')
    hadoop_conf.set('fs.s3a.block.size', '512M')
    hadoop_conf.set('fs.s3a.committer.magic.enabled', 'false')
    hadoop_conf.set('fs.s3a.committer.name', 'directory')
    hadoop_conf.set('fs.s3a.committer.staging.abort.pending.uploads', 'true')
    hadoop_conf.set('fs.s3a.committer.staging.conflict-mode', 'append')
    hadoop_conf.set('fs.s3a.committer.staging.tmp.path', '/tmp/staging')
    hadoop_conf.set('fs.s3a.committer.staging.unique-filenames', 'true')
    hadoop_conf.set('fs.s3a.committer.threads', '2048')
    hadoop_conf.set('fs.s3a.connection.establish.timeout', '5000')
    hadoop_conf.set('fs.s3a.connection.maximum', '8192')
    hadoop_conf.set('fs.s3a.connection.ssl.enabled', 'false')
    hadoop_conf.set('fs.s3a.connection.timeout', '200000')
    hadoop_conf.set('fs.s3a.fast.upload.active.blocks', '2048')
    hadoop_conf.set('fs.s3a.fast.upload.buffer', 'disk')
    hadoop_conf.set('fs.s3a.max.total.tasks', '2048')
    hadoop_conf.set('fs.s3a.multipart.size', '512M')
    hadoop_conf.set('fs.s3a.multipart.threshold', '512M')
    hadoop_conf.set('fs.s3a.socket.recv.buffer', '65536')
    hadoop_conf.set('fs.s3a.socket.send.buffer', '65536')
    hadoop_conf.set('fs.s3a.threads.max', '2048')

    s3_max_connection = os.getenv('S3_MAX_CONNECTION', '200')
    # one thread handles 10 connections at the same time
    s3_max_thread_writting = int(int(s3_max_connection) / 10)
    hadoop_conf.set('fs.s3a.connection.maximum', s3_max_connection)
    hadoop_conf.set("fs.s3a.threads.max", str(s3_max_thread_writting))

def register_spark_task(spark=None):
    from pyspark.sql.session import SparkSession
    from aiaengine.pipeline.providers.spark import (
        PlatformSparkDatasetWriter,
        SparkDatasetReader,
        SparkDatasetWriterProvider,
        SparkDatasetReaderProvider,
        SparkSessionProvider
    )

    client = Client()

    if spark is None:
        spark = SparkSession.builder.appName(
            os.getenv('JOB_ID', "PipelineTaskApp")).getOrCreate()
        set_hadoop_spark_parameters(spark)

    register_converters(client, spark, None, None)


    clusters = {
        CLUSTER_TYPE_NONE: None,
        CLUSTER_TYPE_SPARK: spark,
    }
    dataset_readers = {
        DATAFRAME_FORMAT_SPARK: SparkDatasetReader(spark=spark),
        DATAFRAME_FORMAT_PANDAS: PandasDatasetReader()
    }
    register_service_providers(client=client,
                               clusters=clusters,
                               dataset_readers=dataset_readers)

    register_service_provider('SparkSession', SparkSessionProvider(spark=spark))
    register_service_provider('SparkDatasetWriter', SparkDatasetWriterProvider(
        writer=PlatformSparkDatasetWriter(client=client, spark=spark)))
    register_service_provider('SparkDatasetReader', SparkDatasetReaderProvider(
        reader=SparkDatasetReader(spark=spark)))

    enable_task_decoration()

    logging.basicConfig(level=logging.INFO)

    return spark, client

def is_spark_env():
    return os.getenv('AIA_ENGINE_PIPELINE_TASK_SPARK_ENABLED', 'false') == 'true'

def is_dask_env():
    return os.getenv('AIA_ENGINE_PIPELINE_TASK_DASK_ENABLED', 'false') == 'true'

def is_cuml_env():
    return os.getenv('AIA_ENGINE_PIPELINE_TASK_CUML_ENABLED', 'false') == 'true'

def is_dask_cuml_env():
    return os.getenv('AIA_ENGINE_PIPELINE_TASK_DASK_CUML_ENABLED', 'false') == 'true'


def register_task(spark=None, dask_cluster=None, dask_cuda_cluster=None):
    client = Client()
    clusters = {
        CLUSTER_TYPE_NONE: None,
    }

    dataset_readers = {
        DATAFRAME_FORMAT_PANDAS: PandasDatasetReader(),
    }
    register_service_provider('PandasDatasetReader', 
        PandasDatasetReaderProvider(reader=dataset_readers[DATAFRAME_FORMAT_PANDAS])
    )

    if is_spark_env():
        from pyspark.sql.session import SparkSession
        from aiaengine.pipeline.providers.spark import (
            PlatformSparkDatasetWriter,
            SparkDatasetReader,
            SparkDatasetWriterProvider,
            SparkDatasetReaderProvider,
            SparkSessionProvider
        )

        if spark is None:
            spark = SparkSession.builder.appName(
                os.getenv('JOB_ID', "PipelineTaskApp")).getOrCreate()
            set_hadoop_spark_parameters(spark)

        clusters[CLUSTER_TYPE_SPARK] = spark
        dataset_readers[DATAFRAME_FORMAT_SPARK] = SparkDatasetReader(spark=spark)

        register_service_provider('SparkSession', SparkSessionProvider(spark=spark))
        register_service_provider('SparkDatasetWriter',
            SparkDatasetWriterProvider(
                writer=PlatformSparkDatasetWriter(client=client, spark=spark)
            )
        )
        register_service_provider('SparkDatasetReader',
            SparkDatasetReaderProvider(
                reader=SparkDatasetReader(spark=spark)
            )
        )
    if is_dask_env():
        from dask.distributed import LocalCluster
        from aiaengine.pipeline.providers.dask import DaskDatasetReader, DaskDatasetReaderProvider

        if dask_cluster is None:
            dask_cluster = LocalCluster()
        clusters[CLUSTER_TYPE_DASK] = dask_cluster
        dataset_readers[DATAFRAME_FORMAT_DASK] = DaskDatasetReader()
        register_service_provider('DaskDatasetReader', 
            DaskDatasetReaderProvider(reader=dataset_readers[DATAFRAME_FORMAT_DASK])
        )

    if is_cuml_env():
        from aiaengine.pipeline.providers.cuml import CumlDatasetReader, CumlDatasetReaderProvider
        dataset_readers[DATAFRAME_FORMAT_CUDF] = CumlDatasetReader()
        register_service_provider('CumlDatasetReader',
            CumlDatasetReaderProvider(reader=dataset_readers[DATAFRAME_FORMAT_CUDF])
        )

    if is_dask_cuml_env():
        from dask_cuda import LocalCUDACluster
        from aiaengine.pipeline.providers.dask_cudf import DaskCudfDatasetReader, DaskCudfDatasetReaderProvider

        if dask_cuda_cluster is None:
            dask_cuda_cluster = LocalCUDACluster()
        clusters[CLUSTER_TYPE_DASK_CUDA] = dask_cuda_cluster
        dataset_readers[DATAFRAME_FORMAT_DASK_CUDF] = DaskCudfDatasetReader()
        register_service_provider('DaskCudfDatasetReader',
            DaskCudfDatasetReaderProvider(reader=dataset_readers[DATAFRAME_FORMAT_DASK_CUDF])
        )

    register_converters(client, spark, dask_cluster, dask_cuda_cluster)
    register_service_providers(client=client,
                               clusters=clusters,
                               dataset_readers=dataset_readers)

    enable_task_decoration()

    logging.basicConfig(level=logging.INFO)