import dataclasses
from typing import List, Optional, Union
import json

from aiaengine.api import data_source as api_data_source
from aiaengine.api import pipeline as api_pipeline
from aiaengine.api import dataset as api_dataset
from aiaengine.api import app as api_app
from aiaengine.api import project as api_project
from aiaengine.api import recipe as api_recipe
from aiaengine.api import user as api_user
from aiaengine.app import (
    AnomalyDetectionConfig,
    ClassificationConfig,
    ClusteringConfig,
    MLProblemType,
    RegressionConfig,
    TimeSeriesForecastingConfig,
)

from google.protobuf import struct_pb2, json_format

from aiaengine.common import DEFAULT_TIMEOUT, STATUS_FAILED, STATUS_SUCCESS, get_default_client
from aiaengine.app import App, SupervisedLearningConfig
from aiaengine.data_source import DataSource
from aiaengine.dataset import Dataset
from aiaengine.common import Resource
from aiaengine.recipe import Recipe
from aiaengine.user import User, UserRoles
from aiaengine.waiter import wait_until
from aiaengine.util import unpack_grpc_error
from grpc import RpcError

@dataclasses.dataclass
class ProjectUser:
    """The object model represents a user in the project"""
    user: User
    """User information"""
    role: str
    """Role of user in the project"""

class Project(Resource):
    """The object model represents a project in the Engine
    """

    @property
    def name(self) -> str:
        """Name of the project

        Returns:
            str: Name of the project
        """
        return self._data.name

    @property
    def description(self) -> str:
        """Description of the project

        Returns:
            str: Description of the project
        """
        return self._data.description

    def update(self, name: str, description: str = None):
        """Update information of the project

        Args:
            name (str): updated project name
            description (str, optional): updated project description. Defaults to None.
        """
        self.client.projects.UpdateProject(
            api_project.UpdateProjectRequest(
                id=self.id,
                name=name,
                description=description
            )
        )

    def delete(self):
        """Delete the project.

            Warning:
                All related data (datasets, recipes, apps) of the project will be deleted.
        """
        self.client.projects.DeleteProject(
            api_project.DeleteProjectRequest(
                id=self.id
            )
        )

    def create_dataset(
        self,
        name: str,
        data_source: DataSource,
        timeout: Optional[int] = DEFAULT_TIMEOUT,
    ):
        """Create a dataset in the project from a data source (File or Database)

        Args:
            project_id (str): ID of the project in which this dataset is imported
            name (str): name of the dataset
            data_source: source of data to be imported
            columns: Schema of the

        Returns:
            str: ID of newly created dataset
        """
        data_source.client = self.client
        data_source.create(self.id, name)

        # calling create datasets
        create_datasets_request = {
            "id": data_source.id,
            "datasets": [
                {
                    "dataset_name": name,
                    "data_item_ids": [
                        data_item.id for data_item in data_source.data.data_items
                    ],
                }
            ],
        }

        response = self.client.datasources.CreateDatasets(
            json_format.ParseDict(
                create_datasets_request, api_data_source.CreateDatasetsRequest()
            )
        )

        job_id = response.job_ids[0]
        new_dataset_id = response.dataset_ids[0]
        print(f'New dataset id={new_dataset_id}')

        def dataset_processed():
            rs = self.client.pipelines.GetResult(
                api_pipeline.GetResultRequest(job_id=job_id)
            )
            if rs.status == STATUS_SUCCESS:
                return self.client.datasets.GetDataset(
                    api_dataset.GetDatasetRequest(id=new_dataset_id)
                )
            elif rs.status == STATUS_FAILED:
                raise RuntimeError(
                    f"Failed to create dataset from DataSource(id={data_source.id}): {rs.error_message}"
                )

        dataset_data = wait_until(dataset_processed, timeout=timeout)
        return Dataset(id=dataset_data.id, data=dataset_data, client=self.client)

    def list_datasets(self) -> List[Dataset]:
        response = self.client.datasets.ListDatasets(
            api_dataset.ListDatasetsRequest(project_id=self.id)
        )
        return [
            Dataset(id=dataset_data.id, data=dataset_data, client=self.client)
            for dataset_data in response.datasets
        ]

    def get_dataset(self, id: str):
        return Dataset(id=id, client=self.client)

    def create_recipe(
        self, name, dataset_id, target_columns: List[str] = None, description: str = ""
    ):
        recipe_data = self.client.recipes.CreateRecipe(
            api_recipe.CreateRecipeRequest(
                name=name,
                description=description,
                datasets=[
                    api_recipe.InputDataset(
                        id=dataset_id, target_columns=target_columns
                    )
                ],
            )
        )
        return Recipe(id=recipe_data.id, data=recipe_data, client=self.client)
    
    def list_recipes(self) -> List[Recipe]:
        response = self.client.recipes.ListRecipes(
            api_recipe.ListRecipesRequest(project_id=self.id)
        )
        return [Recipe(id=r.id, data=r, client=self.client) for r in response.recipes]

    def create_app(
        self,
        name: str,
        dataset_id: str,
        config: Union[
            ClassificationConfig,
            RegressionConfig,
            ClusteringConfig,
            AnomalyDetectionConfig,
            TimeSeriesForecastingConfig,
        ],
        description: str = "",
        timeout=DEFAULT_TIMEOUT,
    ):
        """Create a new ML application

        Args:
            name (str): Name of the application
            dataset_id (str): ID of the input dataset
            config (Union[ ClassificationConfig, RegressionConfig, ClusteringConfig ]): Configuration for the problem type
            training_data_proportion (float, optional): train/test split ratio. Deprecated. Use `train_data_propotion` of . Defaults to 0.8.
            description (str, optional): Descripton of the application. Defaults to "".
            timeout (int, optional): waiting time for the app to be processed. Defaults to DEFAULT_TIMEOUT.

        Raises:
            Exception: _description_

        Returns:
            App: _description_
        """
        target_column = None
        if isinstance(config, SupervisedLearningConfig):
            target_column = config.target_column
            if target_column is None:
                raise Exception("target_column is required")
            elif not isinstance(target_column, str) or len(target_column) == 0:
                raise Exception(
                    "Invalid target_column. It must be a str and not be empty."
                )
        problem_type = None
        if isinstance(config, ClassificationConfig):
            problem_type = MLProblemType.CLASSIFICATION
        elif isinstance(config, RegressionConfig):
            problem_type = MLProblemType.REGRESSION
        elif isinstance(config, ClusteringConfig):
            problem_type = MLProblemType.CLUSTERING
        elif isinstance(config, TimeSeriesForecastingConfig):
            problem_type = MLProblemType.TIME_SERIES_FORECASTING
        elif isinstance(config, AnomalyDetectionConfig):
            problem_type = MLProblemType.ANOMALY_DETECTION
        
        if problem_type is None:
            raise Exception("""`config` must be provided with appropriate configuration class:
                [ClassificationConfig, Regression, ClusteringConfig, TimeSeriesForecastingConfig, AnomalyDetectionConfig]""")

        api_problem_type_config = struct_pb2.Struct()
        if config is not None:
            json_format.ParseDict(
                json.loads(config.to_json()), api_problem_type_config # json.loads(x.to_json()) is a workaround for serializing enum values
            )
        target_columns = []
        if target_column is not None:
            target_columns.append(target_column)
        app_data = self.client.apps.CreateApp(
            api_app.CreateAppRequest(
                name=name,
                description=description,
                dataset_id=dataset_id,
                target_columns=target_columns,
                problem_type=problem_type,
                problem_type_config=api_problem_type_config,
                extra_columns={},
                training_data_proportion=0,
            )
        )
        app_id = app_data.id

        def app_processed():
            response = self.client.apps.GetApp(api_app.GetAppRequest(id=app_id))
            if response.status == STATUS_SUCCESS:
                return response
            elif response.status == STATUS_FAILED:
                raise RuntimeError(f"Failed to process App(id={app_id})")

        app_data = wait_until(app_processed, timeout=timeout)

        return App(id=app_id, data=app_data, client=self.client)

    def list_apps(self) -> List[App]:
        """Get all apps in the project

        Returns:
            List[App]: List of applications in the projects
        """
        response = self.client.apps.ListApps(
            api_app.ListAppsRequest(
                project_id=self.id
            )
        )
        return [
            App(id=app.id, data=app, client=self.client)
            for app in response.apps
        ]

    def add_user(self, user_email: str, role: str = UserRoles.ProjectViewer):
        """Add an existing user to the project

        Args:
            user_id (str): Unique identity of the added user
            role (str, optional): Role of the user in the project. Defaults to UserRoles.ProjectViewer.
        """
        try:
            user = self.client.users.GetUserByEmail(api_user.GetUserByEmailRequest(
                email=user_email
            ))
            self.client.projects.CreateProjectUser(api_project.CreateProjectUserRequest(
                id=self.id,
                user_id=user.id,
                role_id=role,
            ))
        except RpcError as e:
            error_message, error_code = unpack_grpc_error(e)
            raise RuntimeError(f"{error_code}: {error_message}")

    def remove_user(self, user_email: str):
        """Remove a user from the project

        Args:
            user_id (str): Unique identity of the removed user
        """
        try:
            user = self.client.users.GetUserByEmail(api_user.GetUserByEmailRequest(
                email=user_email
            ))
            self.client.projects.DeleteProjectUser(api_project.DeleteProjectUserRequest(
                id=self.id,
                user_id=user.id,
            ))
        except RpcError as e:
            error_message, error_code = unpack_grpc_error(e)
            raise RuntimeError(f"{error_code}: {error_message}")

    def list_users(self):
        """Return all users of the project

        Returns:
            List of project users
        """
        response = self.client.projects.ListProjectUsers(
            api_project.ListProjectUsersRequest(id=self.id)
        )

        return [
            ProjectUser(
                user=User(
                    id=project_user.user.id,
                    name=project_user.user.display_name,
                    email=project_user.user.email
                ), 
                role=project_user.role
            )
            for project_user in response.project_users
        ]

    @staticmethod
    def list_user_projects():
        """Return all projects which can be accessed by the current user

        Returns:
            List of projects which can be accessed by the current user
        """
        client = get_default_client()
        response = client.projects.ListUserProjects(
            api_project.ListUserProjectsRequest()
        )
        return [
            Project(id=project.id, data=project, client=client)
            for project in response.projects
        ]

    def _get_data(self) -> any:
        return self.client.projects.GetProject(
            api_project.GetProjectRequest(id=self.id)
        )

    def __repr__(self) -> str:
        return f"Project(id='{self.id}', name='{self.name}', description='{self.description}')"
