from typing import List, NewType
import requests
import json

from aiaengine.waiter import wait_until
from aiaengine.api import recipe as api_recipe
from aiaengine.api.client import Client
from aiaengine.common import Resource
from aiaengine.dataset import Dataset
from aiaengine.core.recipe import PRIMARY_DATASET_ALIAS
from aiaengine.waiter import wait_until, wait_pipeline_run
from aiaengine.common import DEFAULT_TIMEOUT

# forward declaration
Recipe = NewType("Recipe", Resource)

class Recipe(Resource):
    """Recipe management
    """
    def __init__(self, id, data=None, client: Client = None):
        super().__init__(id, data, client)

        if data is None:
            request = api_recipe.GetRecipeRequest(id=id)
            data = self.client.recipes.GetRecipe(request)
            self.id = data.id
        self._data = data
        self.name = data.name
        self.current_iteration = len(self._data.iterations)
        self.target_columns = []

    def update(self, name: str, descriptpion: str = ''):
        self.client.recipes.UpdateRecipe(
            api_recipe.UpdateRecipeRequest(
                id=self.id,
                name=name,
                descriptpion=descriptpion
            )
        )

    def delete(self):
        self.client.recipes.DeleteRecipe(
            api_recipe.DeleteRecipeRequest(
                id=self.id
            )
        )

    def get_recommendations(self, timeout=300):
        """Get recommended actions

        Args:
            timeout (int, optional): waiting time in seconds. Defaults to 300.

        Returns:
            _type_: _description_
        """

        # wait until the recommendations are available
        def recommendation_processed():
            return self._wait_for_recommendation_status(status="success")

        wait_until(recommendation_processed, timeout=timeout)
        request = api_recipe.GetRecommendedActionsUrlRequest(
            id=self.id, iteration=self.current_iteration
        )
        response = self.client.recipes.GetRecommendedActionsUrl(request)
        if response.url != "":
            r = requests.get(url=response.url)
            if r.status_code != 200:
                return []
            return r.json()

    def add_recommended_actions(self, recommendation):
        """Add actions from a recommendation

        Args:
            recommendation (_type_): _description_
        """
        self._add_actions_to_recipe(
            recipe_id=self.id,
            iteration=self.current_iteration,
            actions=recommendation["solution"]["actions"],
        )

    def add_input_dataset(self, dataset_id: str, alias: str):
        request = api_recipe.UpdateInputDatasetsRequest(
            id=self.id,
            datasets=[
                api_recipe.InputDataset(
                    id=dataset_id,
                    alias=alias,
                )
            ],
            override=False,
        )
        self.client.recipes.UpdateInputDatasets(request)

    def add_actions(self, actions):
        """Add new actions to the recipe"""
        return self._add_actions_to_recipe(
            recipe_id=self.id,
            iteration=self.current_iteration,
            actions=actions
        )

    def set_target_columns(self, target_columns: List[str]):
        """Set the target column for the recipe. This will assist the recommendation agents

        Args:
            target_columns (str): Name of the target columns in the dataset
        """
        self.target_columns = target_columns

    def commit_actions(self, pipeline_configuration_id=None, timeout=300):
        """Commit actions of the current iteration

        Args:
            pipeline_configuration_id (str, optional): _description_. Defaults to "".
            timeout (int, optional): waiting time in seconds. Defaults to 300.

        Returns:
            _type_: _description_
        """
        commit_actions_request = api_recipe.CommitActionsRequest(
            id=self.id,
            iteration=self.current_iteration,
            target_columns=self.target_columns,
            pipeline_configuration_id=pipeline_configuration_id,
        )
        self.client.recipes.CommitActions(commit_actions_request)

        if timeout > 0:
            # wait until the actions are committed successfully
            def actions_processed():
                return self._wait_for_status(step="file_format", status="success")

            wait_until(actions_processed, timeout=timeout)
            self.current_iteration = len(self._data.iterations)

    def complete(self, output_dataset_name, timeout=300) -> Dataset:
        """Complete the recipe

        Args:
            output_dataset_name (_type_): name of the output dataset
            timeout (int, optional): waiting time in seconds. Defaults to 300.

        Raises:
            Exception: _description_

        Returns:
            _type_: _description_
        """
        if output_dataset_name is None or output_dataset_name.strip() == "":
            raise Exception("Output dataset name is required")
        request = api_recipe.CompleteRecipeRequest(
            id=self.id, dataset_name=output_dataset_name
        )
        response = self.client.recipes.CompleteRecipe(request)
        output_dataset = Dataset(id=response.dataset_id, client=self.client)
        return output_dataset

    def apply(self, datasets: dict, output_dataset_name: str, timeout: int = DEFAULT_TIMEOUT) -> Dataset:
        """Apply the recipe with new input datasets

        Args:
            datasets (dict): input datasets: alias <-> ID
            output_dataset_name (str): name of output dataset
            timeout (int, optional): time to wait for the recipe to be applied. Defaults to DEFAULT_TIMEOUT.

        Returns:
            output dataset
        """
        response = self.client.recipes.RunRecipe(api_recipe.RunRecipeRequest(
            id=self.id,
            dataset_id=datasets[PRIMARY_DATASET_ALIAS],
            new_dataset_name=output_dataset_name
        ))
        pipeline_run_id = response.job_id
        pr = wait_pipeline_run(self.client, pipeline_run_id=pipeline_run_id, timeout=timeout)
        output_dataset_id = pr.results[0].result['new_dataset_id']
        return Dataset(id=output_dataset_id, client=self.client)

    def copy(self, input_datasets: dict, name: str, description: str = '') -> Recipe:
        """Create a copy of the current dataset

        Args:
            input_datasets (dict): Input datasets
            name (str): name of the new recipe
            description (str, optional): description of the new recipe. Defaults to ''.

        Returns:
            Recipe: new recipe
        """

        # validate required input datasets
        for ds in self._data.datasets:
            if ds.alias not in input_datasets:
                raise Exception(f"Missing input dataset with alias '{ds.alias}'")

        new_recipe_data = self.client.recipes.CreateRecipe(
            api_recipe.CreateRecipeRequest(
                name=name,
                description=description,
                datasets=[
                    api_recipe.InputDataset(
                        id=input_datasets[PRIMARY_DATASET_ALIAS], target_columns=self.target_columns
                    )
                ],
            )
        )

        # copy datasets
        request = api_recipe.UpdateInputDatasetsRequest(
            id=new_recipe_data.id,
            datasets=[
                api_recipe.InputDataset(
                    id=dataset_id,
                    alias=alias,
                )
                for (alias, dataset_id) in input_datasets.items()
                if alias != PRIMARY_DATASET_ALIAS
            ],
            override=False,
        )
        self.client.recipes.UpdateInputDatasets(request)

        # copy actions
        actions = []
        for it in self._data.iterations:
            response = self.client.recipes.GetCommittedActionsUrl(api_recipe.GetCommittedActionsUrlRequest(id=self.id, iteration=it.iteration))
            if (response.url != ''):
                data = requests.get(response.url).json()
                actions = actions + data[0]['solution']['actions']

        new_recipe = Recipe(id=new_recipe_data.id, client=self.client)
        result = new_recipe.add_actions(actions)
        if result['invalid_index'] != -1:
            raise Exception("Incompatible input datasets provided: ")

        return new_recipe

    def _get_recipe_data(self):
        request = api_recipe.GetRecipeRequest(id=self.id)
        self._data = self.client.recipes.GetRecipe(request)

    def _wait_for_status(self, step, status):
        self._get_recipe_data()
        iteration = self._data.iterations[self.current_iteration - 1]
        return iteration.step == step and iteration.status == status

    def _wait_for_recommendation_status(self, status):
        self._get_recipe_data()
        iteration = self._data.iterations[self.current_iteration - 1]
        return iteration.recommendation_status == status

    def _add_actions_to_recipe(self, recipe_id, actions, iteration=1):
        request = api_recipe.AddActionsRequest(
            id=recipe_id, iteration=iteration, actions=json.dumps(actions, default=vars)
        )
        response = self.client.recipes.AddActions(request)
        if response.invalid_index != -1:
            return {
                "invalid_index": response.invalid_index,
                "error": json.loads(response.error),
            }
        else:
            return {
                "invalid_index": response.invalid_index
            }
