import base64
from typing import List, Dict, Tuple
import json
import os
import tarfile
import tempfile
from urllib.parse import urlparse
from urllib.request import urlopen
import boto3
from shutil import copyfile
import logging
from tqdm import tqdm, trange
from time import sleep
import logging

from aiaengine import api
from aiaengine.api import file
from aiaengine.api import dataset_pb2 as dataset
from aiaengine.api import pipeline
from google.protobuf import json_format

DATASET = "dataset"
APP = "app"
MODEL = "model"
TRAINING = "training"
FEATURE_SET = "featureset"
EVALUATION = "evaluation"

def get_dataset_file_urls(client, dataset_id):
    # get list of dataset's files
    list_files_response = client.files.ListFiles(
        file.ListFilesRequest(dataset_id=dataset_id)
    )
    return [f.url for f in list_files_response.files]


def get_dataset_file_download_urls(client, dataset_id):
    dataset_info = client.datasets.GetDataset(dataset.GetDatasetRequest(id=dataset_id))
    # get list of dataset's files
    list_files_response = client.files.ListFiles(
        file.ListFilesRequest(dataset_id=dataset_id)
    )
    file_urls = []
    for f in list_files_response.files:
        r = client.files.GetFileDownloadUrl(file.GetFileDownloadUrlRequest(id=f.id, project_id=dataset_info.project.id))
        file_urls.append(r.url)
    return file_urls


def _get_s3():
    ENDPOINT = os.environ["AWS_ENDPOINT"]
    if not ENDPOINT.startswith("http"):
        ENDPOINT = "https://" + ENDPOINT
    MINIO_ACCESS_KEY = os.environ["AWS_ACCESS_KEY_ID"]
    MINIO_SECRET_KEY = os.environ["AWS_SECRET_ACCESS_KEY"]

    logging.debug(f"s3 info: {ENDPOINT}, {MINIO_ACCESS_KEY}")
    return boto3.resource(
        "s3",
        endpoint_url=ENDPOINT,
        aws_access_key_id=MINIO_ACCESS_KEY,
        aws_secret_access_key=MINIO_SECRET_KEY,
    )


def download_file(url, path):
    """Download file from S3 or local file to 'path'
    param: @url
    paeram: @path
    """
    # just do a copy if URL is local
    if url.startswith("/") or url.startswith("file://"):
        copyfile(url, path)
    else:
        s3 = _get_s3()
        o = urlparse(url)
        s3.Object(o.netloc, o.path.lstrip("/")).download_file(path)


def upload_file(path, url):
    """Upload file from local file system to S3"""
    if os.path.exists(path) == False:
        logging.error("File not found {}".format(path))
        return
    s3 = _get_s3()
    o = urlparse(url)
    logging.info(f"Uploading {path} to {url}")
    s3.Object(o.netloc, o.path.lstrip("/")).upload_file(path)


def download_model(url, path):
    (_, compressed_model_path) = tempfile.mkstemp()
    download_file(url, compressed_model_path)

    tar = tarfile.open(compressed_model_path, "r:gz")
    logging.debug("Extract data to " + path)
    tar.extractall(path)
    tar.close()
    logging.debug("Done")


def upload_model(path, url):
    logging.info("Upload model from " + path + " to " + url)
    logging.info("Compressing model...")
    (_, compressed_model_path) = tempfile.mkstemp()
    tar = tarfile.open(compressed_model_path, "w:gz")
    for name in os.listdir(path):
        logging.info(f"Compressing file {name}")
        tar.add(os.path.join(path, name), arcname=name)

    # for cur_path, _, files in os.walk(path):
    #     for file in files:
    #         __local_file = os.path.join(cur_path, file)
    #         logging.debug("Add", __local_file, "to tarfile")
    #         tar.add(__local_file, arcname=name)
    tar.close()
    upload_file(compressed_model_path, url)


def list_files_in_folder(path, suffix):
    """
    path: s3://bucket/folder
    suffix: .csv or .parquet
    """
    path_elements = path.split("/")
    bucket = path_elements[2]
    prefix = "/".join(path_elements[3:])
    if not prefix.endswith("/"):
        prefix = prefix + "/"

    return [
        "s3://{}/{}".format(bucket, obj.key)
        for obj in _get_s3().Bucket(bucket).objects.filter(Prefix=prefix)
        if obj.key.endswith(suffix)
    ]


def list_local_files_in_folder(path, suffix):
    if os.path.isfile(path):
        if path.endswith(suffix):
            return [path]
        else:
            return []

    files = []
    for dirpath, _, filenames in os.walk(path):
        files.extend(
            [os.path.join(dirpath, f) for f in filenames if f.endswith(suffix)]
        )

    return files


def parse_params(params=""):
    params = json.loads(base64.b64decode(params))

    return params


def save_to_file(base64_encoded_content, local_path):
    decoded = base64.b64decode(base64_encoded_content)
    with open(local_path, "w+b") as output_file:
        output_file.write(decoded)


def download_s3_presigned_url(url, local_path):
    with open(local_path, "w") as f:
        r = urlopen(url)
        f.write(r.read().decode("utf-8"))


def download_metadata_of_dataset(client, dataset_id, local_path):
    metadata_resp = client.datasets.GetMetadataUrl(
        dataset.GetMetadataUrlRequest(id=dataset_id)
    )
    download_s3_presigned_url(metadata_resp.url, local_path)


def download_schema_of_dataset(client, dataset_id, local_path):
    schema_resp = client.datasets.GetSchemaUrl(
        dataset.GetSchemaUrlRequest(id=dataset_id)
    )
    download_s3_presigned_url(schema_resp.url, local_path)


def report_progress(task_name, progress, api_client, pipelinerun_id):
    if api_client != None:
        data = {
            "id": pipelinerun_id,
            "tasks": [{"name": task_name, "progress": progress}],
        }
        request = json_format.Parse(json.dumps(data), pipeline.UpdateProgressRequest())
        logging.info(f'Report progress of pipeline run = {pipelinerun_id}, task = "{task_name}", progress = {progress}')
        api_client.pipelines.UpdateProgress(request)


def report_progress_async(task_name, progress, api_client, pipelinerun_id):
    if api_client != None:
        data = {
            "id": pipelinerun_id,
            "tasks": [{"name": task_name, "progress": progress}],
        }
        request = json_format.Parse(json.dumps(data), pipeline.UpdateProgressRequest())
        logging.info(f'Report progress of pipeline run = {pipelinerun_id}, task = "{task_name}", progress = {progress}')
        api_client.pipelines.UpdateProgress.future(request)
        


def report_pipeline_status(success, api_client, pipelinerun_id):
    # report status to pipelineRun
    if api_client != None:
        status = 1 if success else 2
        api_client.pipelines.UpdateProgress(
            pipeline.UpdateProgressRequest(id=pipelinerun_id, status=status)
        )
        logging.info(f"Report pipeline status: {status}")


def download_model_templates(api_client, path):
    resp = api_client.templates.ListTemplates(api.Empty())
    json_str = json_format.MessageToJson(resp)
    with open(path, "wb") as f:
        f.write(json_str.encode("utf-8"))


def get_progress(client, entity_type, entity_id):
    metadata = {
        "dataset": {"actionName": "system.action.data-analysis"},
        "app": {"actionName": "system.action.create-app"},
        "featureset": {"actionName": "system.action.model-recommendation"},
        "model": {"actionName": "system.action.model-training"},
    }
    get_progress_request = pipeline.GetProgressRequest(
        owner_id=entity_id, owner_type=entity_type, metadata=metadata[entity_type]
    )
    get_progress_response = client.pipelines.GetProgress(get_progress_request)
    return round(get_progress_response.progress)



def wait_for_processing(client, entity_type, entity_id, timeout=300):
    with tqdm(total=100) as pbar:
        while pbar.n < 100 and timeout > 0:
            current_progress = get_progress(
                client, entity_type=entity_type, entity_id=entity_id
            )
            pbar.update(current_progress - pbar.n)
            sleep(1)
            timeout -= 1
        if timeout <= 0:
            raise Exception(
                f"time out when waiting for {entity_type}({entity_id}) processed"
            )


def write_argo_output_parameter(parameter_name, str_value):
    with open(f"/tmp/{parameter_name}.txt", "w") as out:
        out.write(str_value)


def report_pipelinerun_result(
    result: dict, step_name=None, job_id=None, status=None, use_sdk=False
):
    """
        Report result or/and status to pipelinerun.

        Parameters:
            job_id: pipelinerun ID. It can be None if use_sdk=False
            result: dict[string,string]. It can be None if you just need report pipeline's status
            step_name: step name of this result. Default is None. If the pipeline has many steps, each step has result, we must provide step name.
            status: 1: success, 2: failed, default is None
            use_sdk: True: use SDK to call API pipeline.UpdateResult(). False: write result to file /tmp/RESULT.txt
    """
    logging.info(
        f"Report pipelinerun result: {result}, step_name={step_name}, status={status}, job_id={job_id}, use_sdk={use_sdk}"
    )
    if status is None:
        # a step is used in many places, such as in continuous pipeline, it need report status for child pipelinerun
        is_report_pipeline_status = (
            os.getenv("IS_REPORT_PIPELINE_STATUS", "false") == "true"
        )
        if is_report_pipeline_status:
            status = 1  # success

    if not use_sdk:
        wrapper_result = {}
        if status is not None:
            wrapper_result["status"] = status
        if step_name is not None:
            wrapper_result["step_name"] = step_name
        if result is not None:
            wrapper_result["result"] = result
        write_argo_output_parameter("RESULT", json.dumps(wrapper_result))
    else:
        client = api.Client()
        req = pipeline.UpdateResultRequest(
            job_id=job_id, status=status, step_name=step_name, result=result
        )
        client.pipelines.UpdateResult(req)


def stop_on_failed(error_message, 
                   error_code=None, 
                   step_name=None, 
                   exit_code=1,
                   **kwargs):
    """
        Write error message to /dev/termination-log and stop process.
        Argo will get message and stop workflow

        Parameters:
            error_message: error message show to user
            error_code: error code
            step_name: step name
            exit_code: exit code of program
    """
    logging.info(f"Step {step_name} failed and stop: {error_message}, error code={error_code}")
    wrote_message = error_message
    if error_code or step_name:
        reported_error = {
            'error_message': error_message,
            'error_code': error_code,
            'step_name': step_name
        }
        if 'error_data' in kwargs:
            reported_error['error_data'] = kwargs['error_data']
        if 'suggestion' in kwargs:
            reported_error['suggestion'] = kwargs['suggestion']
        wrote_message = json.dumps(reported_error)
    file_path = "/dev/termination-log"
    with open(file_path, "w") as out:
        out.write(wrote_message)
    import os
    os._exit(exit_code)


def generate_file_upload_url(
    client, project_id, name="data.parquet", content_type="application/binary+parquet"
):
    """ Generate output URL in platform """
    logging.info(f"Generate file upload url in project {project_id}")
    request = file.UploadFileRequest(
        name=name, content_type=content_type, project_id=project_id
    )
    response = client.files.UploadFile(request=request)
    return response.url


def unpack_grpc_error(grpc_err):
    from google.rpc.error_details_pb2 import ErrorInfo
    from grpc_status import rpc_status

    error_code = "INTERNAL_ERROR"
    error_message = grpc_err.details()
    status = rpc_status.from_call(grpc_err)
    if status is not None:
        logging.debug(status.details)
        detail = status.details[0]
        if detail.Is(ErrorInfo.DESCRIPTOR):
            info = ErrorInfo()
            detail.Unpack(info)
            logging.debug(info)
            error_code = info.reason
    return error_message, error_code

def build_pipeline_run_errors(pipeline_run):
    message = 'Errors:'
    error_steps = [s for s in pipeline_run.results if s.error_message != '']
    for step_result in error_steps:
        message += f"\n-{step_result.error_message}"
    return message
