from tqdm import trange
from time import sleep
from aiaengine.api import Client, model, dataset, featureset, pipeline
from aiaengine.util import build_pipeline_run_errors
from aiaengine.common import(
    LATEST_VERSION, STATUS_SUCCESS, STATUS_FAILED,
    PIPELINE_RUN_STATUS_SUCCESS, PIPELINE_RUN_STATUS_FAILED
)


class FailedEntityError(Exception):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)

def wait_until(task, timeout=300, step=2):
    with trange(
        timeout,
        desc=f"Waiting for {task.__name__}",
        bar_format="{desc} |{bar}| {n_fmt}/{total_fmt}s",
    ) as t:
        for _ in range(0, timeout, step):
            result = task()
            if result:
                return result
            sleep(step)
            t.update(n=step)
    raise TimeoutError(f"Timeout while waiting for task: {task.__name__}")


def get_training(client: Client, model_id, dataset_version: int):
    """Get the training on dataset version

    Args:
        dataset_version (int, optional): version of the training dataset. Defaults to LATEST_VERSION.
    """
    response = client.models.ListTrainings(
        model.ListTrainingsRequest(id=model_id)
    )
    training = None
    latest_version = -1
    for t in response.trainings:
        if (
            dataset_version != LATEST_VERSION
            and t.dataset_version == dataset_version
        ):
            training = t
            break
        if dataset_version == LATEST_VERSION and latest_version < t.dataset_version:
            latest_version = t.dataset_version
            training = t

    if training is None:
        raise RuntimeError(f'Model {model_id} is not trained on dataset version {dataset_version}')

    return training


def wait_model_trained(client: Client, model_id: str, dataset_version: int = LATEST_VERSION, timeout=1200) -> object:
    training = get_training(client, model_id, dataset_version)

    def model_trained():
        last_training = client.models.GetTraining(
            model.GetTrainingRequest(id=model_id, training_id=training.id)
        )
        if last_training.status == STATUS_SUCCESS:
            return last_training
        if last_training.status == STATUS_FAILED:
            raise FailedEntityError(f"Failed to train model {model_id}, training {training.id}")
    return wait_until(model_trained, timeout=timeout, step=3)


def wait_evaluation_processed(client: Client, model_id: str, evaluation_id: str, timeout=600)->object:
    def evaluation_processed():
        e = client.models.GetEvaluation(
            model.GetEvaluationRequest(id=evaluation_id, model_id=model_id)
        )
        if e.status == STATUS_SUCCESS:
            return e
        elif e.status == STATUS_FAILED:
            raise FailedEntityError(f"Failed to evaluate Evaluation(id={evaluation_id}) of Model(id={model_id})")
    return wait_until(evaluation_processed, timeout=timeout)


def wait_featureset_processed(client: Client, feature_set_id:str, timeout=600)->object:
    def feature_set_processed():
        response = client.featuresets.GetFeatureSet(
            featureset.GetFeatureSetRequest(id=feature_set_id)
        )
        if response.status == STATUS_SUCCESS:
            return response
        elif response.status == STATUS_FAILED:
            raise FailedEntityError(f"Failed to process FeatureSet(id={feature_set_id})")
    return wait_until(feature_set_processed, timeout=timeout)


def wait_dataset_processed(client: Client, dataset_id:str, timeout=600)->object:
    def dataset_processed():
        rs = client.datasets.GetDataset(dataset.GetDatasetRequest(id=dataset_id))
        if rs.status == STATUS_SUCCESS:
            return rs
        elif rs.status == STATUS_FAILED:
            raise FailedEntityError(f"Dataset {dataset_id} failed")
    return wait_until(dataset_processed, timeout=timeout)


def wait_pipeline_run(client:Client, pipeline_run_id: str, timeout=600) -> object:
    def pipeline_run_processed():
        pr = client.pipelines.GetPipelineRun(
            pipeline.GetPipelineRunRequest(id=pipeline_run_id)
        )
        if pr.status == PIPELINE_RUN_STATUS_SUCCESS:
            return pr
        elif pr.status == PIPELINE_RUN_STATUS_FAILED:
            raise FailedEntityError(build_pipeline_run_errors(pr))
    return wait_until(pipeline_run_processed, timeout=timeout)
